"""
from from https://github.com/huangtinglin/MixGCF

Created on October 1, 2020

@author: Tinglin Huang (huangtinglin@outlook.com)
"""
import torch
import torch.nn as nn

from cogdl.models import BaseModel


class GraphConv(nn.Module):
    """
    Graph Convolutional Network
    """

    def __init__(self, n_hops, n_users, interact_mat, edge_dropout_rate=0.5, mess_dropout_rate=0.1):
        super(GraphConv, self).__init__()

        self.interact_mat = interact_mat
        self.n_users = n_users
        self.n_hops = n_hops
        self.edge_dropout_rate = edge_dropout_rate
        self.mess_dropout_rate = mess_dropout_rate

        self.dropout = nn.Dropout(p=mess_dropout_rate)  # mess dropout

    def to(self, *args, **kwargs):
        self = super().to(*args, **kwargs)
        self.interact_mat = self.interact_mat.to(*args, **kwargs)
        return self

    def _sparse_dropout(self, x, rate=0.5):
        noise_shape = x._nnz()

        random_tensor = rate
        random_tensor += torch.rand(noise_shape).to(x.device)
        dropout_mask = torch.floor(random_tensor).type(torch.bool)
        i = x._indices()
        v = x._values()

        i = i[:, dropout_mask]
        v = v[dropout_mask]

        out = torch.sparse.FloatTensor(i, v, x.shape).to(x.device)
        return out * (1.0 / (1 - rate))

    def forward(self, user_embed, item_embed, mess_dropout=True, edge_dropout=True):
        # user_embed: [n_users, channel]
        # item_embed: [n_items, channel]

        # all_embed: [n_users+n_items, channel]
        all_embed = torch.cat([user_embed, item_embed], dim=0)
        agg_embed = all_embed
        embs = [all_embed]

        for hop in range(self.n_hops):
            interact_mat = (
                self._sparse_dropout(self.interact_mat, self.edge_dropout_rate) if edge_dropout else self.interact_mat
            )

            agg_embed = torch.sparse.mm(interact_mat, agg_embed)
            if mess_dropout:
                agg_embed = self.dropout(agg_embed)
            # agg_embed = F.normalize(agg_embed)
            embs.append(agg_embed)
        embs = torch.stack(embs, dim=1)  # [n_entity, n_hops+1, emb_size]
        return embs[: self.n_users, :], embs[self.n_users :, :]


class LightGCN(BaseModel):
    @staticmethod
    def add_args(parser):
        """Add model-specific arguments to the parser."""
        # fmt: off
        parser.add_argument('--dim', type=int, default=64, help='embedding size')
        parser.add_argument('--l2', type=float, default=1e-4, help='l2 regularization weight, 1e-5 for NGCF')
        parser.add_argument("--mess_dropout", type=bool, default=False, help="consider mess dropout or not")
        parser.add_argument("--mess_dropout_rate", type=float, default=0.1, help="ratio of mess dropout")
        parser.add_argument("--edge_dropout", type=bool, default=False, help="consider edge dropout or not")
        parser.add_argument("--edge_dropout_rate", type=float, default=0.1, help="ratio of edge sampling")
        parser.add_argument("--ns", type=str, default='mixgcf', help="rns,mixgcf")
        parser.add_argument("--K", type=int, default=1, help="number of negative in K-pair loss")
        parser.add_argument("--n_negs", type=int, default=64, help="number of candidate negative")
        parser.add_argument("--pool", type=str, default='mean', help="[concat, mean, sum, final]")
        parser.add_argument("--context_hops", type=int, default=3, help="hop")
        # fmt: on

    @classmethod
    def build_model_from_args(cls, args):
        return cls(
            args.n_users,
            args.n_items,
            args.l2,
            args.dim,
            args.context_hops,
            args.mess_dropout,
            args.mess_dropout_rate,
            args.edge_dropout,
            args.edge_dropout_rate,
            args.pool,
            args.n_negs,
            args.ns,
            args.K,
            args.adj_mat,
        )

    def __init__(
        self,
        n_users,
        n_items,
        l2,
        dim,
        context_hops,
        mess_dropout,
        mess_dropout_rate,
        edge_dropout,
        edge_dropout_rate,
        pool,
        n_negs,
        ns,
        K,
        adj_mat,
    ):
        super(LightGCN, self).__init__()

        self.n_users = n_users
        self.n_items = n_items
        self.adj_mat = adj_mat

        self.decay = l2
        self.emb_size = dim
        self.context_hops = context_hops
        self.mess_dropout = mess_dropout
        self.mess_dropout_rate = mess_dropout_rate
        self.edge_dropout = edge_dropout
        self.edge_dropout_rate = edge_dropout_rate
        self.pool = pool
        self.n_negs = n_negs
        self.ns = ns
        self.K = K

        self._init_weight()
        self.user_embed = nn.Parameter(self.user_embed)
        self.item_embed = nn.Parameter(self.item_embed)

        self.gcn = self._init_model()

    def _init_weight(self):
        initializer = nn.init.xavier_uniform_
        self.user_embed = initializer(torch.empty(self.n_users, self.emb_size))
        self.item_embed = initializer(torch.empty(self.n_items, self.emb_size))

        # [n_users+n_items, n_users+n_items]
        self.sparse_norm_adj = self._convert_sp_mat_to_sp_tensor(self.adj_mat)

    def _init_model(self):
        return GraphConv(
            n_hops=self.context_hops,
            n_users=self.n_users,
            interact_mat=self.sparse_norm_adj,
            edge_dropout_rate=self.edge_dropout_rate,
            mess_dropout_rate=self.mess_dropout_rate,
        )

    def _convert_sp_mat_to_sp_tensor(self, X):
        coo = X.tocoo()
        i = torch.LongTensor([coo.row, coo.col])
        v = torch.from_numpy(coo.data).float()
        return torch.sparse.FloatTensor(i, v, coo.shape)

    def to(self, *args, **kwargs):
        self = super().to(*args, **kwargs)
        # self.sparse_norm_adj = self.sparse_norm_adj.to(*args, **kwargs)
        self.gcn.to(*args, **kwargs)
        return self

    def forward(self, batch=None):
        user = batch["users"]
        pos_item = batch["pos_items"]
        neg_item = batch["neg_items"]  # [batch_size, n_negs * K]

        # user_gcn_emb: [n_users, channel]
        # item_gcn_emb: [n_users, channel]
        user_gcn_emb, item_gcn_emb = self.gcn(
            self.user_embed, self.item_embed, edge_dropout=self.edge_dropout, mess_dropout=self.mess_dropout
        )

        if self.ns == "rns":  # n_negs = 1
            neg_gcn_embs = item_gcn_emb[neg_item[:, : self.K]]
        else:
            neg_gcn_embs = []
            for k in range(self.K):
                neg_gcn_embs.append(
                    self.negative_sampling(
                        user_gcn_emb, item_gcn_emb, user, neg_item[:, k * self.n_negs : (k + 1) * self.n_negs], pos_item
                    )
                )
            neg_gcn_embs = torch.stack(neg_gcn_embs, dim=1)

        return self.create_bpr_loss(user_gcn_emb[user], item_gcn_emb[pos_item], neg_gcn_embs)

    def negative_sampling(self, user_gcn_emb, item_gcn_emb, user, neg_candidates, pos_item):
        batch_size = user.shape[0]
        s_e, p_e = user_gcn_emb[user], item_gcn_emb[pos_item]  # [batch_size, n_hops+1, channel]
        if self.pool != "concat":
            s_e = self.pooling(s_e).unsqueeze(dim=1)

        """positive mixing"""
        seed = torch.rand(batch_size, 1, p_e.shape[1], 1).to(p_e.device)  # (0, 1)
        n_e = item_gcn_emb[neg_candidates]  # [batch_size, n_negs, n_hops, channel]
        n_e_ = seed * p_e.unsqueeze(dim=1) + (1 - seed) * n_e  # mixing

        """hop mixing"""
        scores = (s_e.unsqueeze(dim=1) * n_e_).sum(dim=-1)  # [batch_size, n_negs, n_hops+1]
        indices = torch.max(scores, dim=1)[1].detach()
        neg_items_emb_ = n_e_.permute([0, 2, 1, 3])  # [batch_size, n_hops+1, n_negs, channel]
        # [batch_size, n_hops+1, channel]
        return neg_items_emb_[[[i] for i in range(batch_size)], range(neg_items_emb_.shape[1]), indices, :]

    def pooling(self, embeddings):
        # [-1, n_hops, channel]
        if self.pool == "mean":
            return embeddings.mean(dim=1)
        elif self.pool == "sum":
            return embeddings.sum(dim=1)
        elif self.pool == "concat":
            return embeddings.view(embeddings.shape[0], -1)
        else:  # final
            return embeddings[:, -1, :]

    def generate(self, split=True):
        user_gcn_emb, item_gcn_emb = self.gcn(self.user_embed, self.item_embed, edge_dropout=False, mess_dropout=False)
        user_gcn_emb, item_gcn_emb = self.pooling(user_gcn_emb), self.pooling(item_gcn_emb)
        if split:
            return user_gcn_emb, item_gcn_emb
        else:
            return torch.cat([user_gcn_emb, item_gcn_emb], dim=0)

    def rating(self, u_g_embeddings=None, i_g_embeddings=None):
        return torch.matmul(u_g_embeddings, i_g_embeddings.t())

    def create_bpr_loss(self, user_gcn_emb, pos_gcn_embs, neg_gcn_embs):
        # user_gcn_emb: [batch_size, n_hops+1, channel]
        # pos_gcn_embs: [batch_size, n_hops+1, channel]
        # neg_gcn_embs: [batch_size, K, n_hops+1, channel]

        batch_size = user_gcn_emb.shape[0]

        u_e = self.pooling(user_gcn_emb)
        pos_e = self.pooling(pos_gcn_embs)
        neg_e = self.pooling(neg_gcn_embs.view(-1, neg_gcn_embs.shape[2], neg_gcn_embs.shape[3])).view(
            batch_size, self.K, -1
        )

        pos_scores = torch.sum(torch.mul(u_e, pos_e), axis=1)
        neg_scores = torch.sum(torch.mul(u_e.unsqueeze(dim=1), neg_e), axis=-1)  # [batch_size, K]

        mf_loss = torch.mean(torch.log(1 + torch.exp(neg_scores - pos_scores.unsqueeze(dim=1)).sum(dim=1)))

        # cul regularizer
        regularize = (
            torch.norm(user_gcn_emb[:, 0, :]) ** 2
            + torch.norm(pos_gcn_embs[:, 0, :]) ** 2
            + torch.norm(neg_gcn_embs[:, :, 0, :]) ** 2
        ) / 2  # take hop=0
        emb_loss = self.decay * regularize / batch_size

        return mf_loss + emb_loss, mf_loss, emb_loss



import numpy as np
import networkx as nx
import scipy.sparse as sp
from sklearn import preprocessing
from sklearn.utils.extmath import randomized_svd
from multiprocessing import Pool
from tqdm import tqdm
import time

from cogdl.utils import alias_draw, alias_setup
from .. import BaseModel


class NetSMF(BaseModel):
    r"""The NetSMF model from the `"NetSMF: Large-Scale Network Embedding as Sparse Matrix Factorization"
    <http://arxiv.org/abs/1710.02971>`_ paper.

    Args:
        hidden_size (int) : The dimension of node representation.
        window_size (int) : The actual context size which is considered in language model.
        negative (int) : The number of nagative samples in negative sampling.
        num_round (int) : The number of round in NetSMF.
        worker (int) : The number of workers for NetSMF.
    """

    @staticmethod
    def add_args(parser):
        """Add model-specific arguments to the parser."""
        # fmt: off
        parser.add_argument("--window-size", type=int, default=10,
                            help="Window size of approximate matrix. Default is 10.")
        parser.add_argument("--negative", type=int, default=1,
                            help="Number of negative node in sampling. Default is 1.")
        parser.add_argument("--num-round", type=int, default=100,
                            help="Number of round in NetSMF. Default is 100.")
        parser.add_argument("--worker", type=int, default=10,
                            help="Number of parallel workers. Default is 10.")
        parser.add_argument("--hidden-size", type=int, default=128)
        # fmt: on

    @classmethod
    def build_model_from_args(cls, args):
        return cls(
            args.hidden_size,
            args.window_size,
            args.negative,
            args.num_round,
            args.worker,
        )

    def __init__(self, dimension, window_size, negative, num_round, worker):
        super(NetSMF, self).__init__()
        self.dimension = dimension
        self.window_size = window_size
        self.negative = negative
        self.worker = worker
        self.num_round = num_round

    def forward(self, graph, return_dict=False):
        self.G = graph.to_networkx()
        node2id = dict([(node, vid) for vid, node in enumerate(self.G.nodes())])
        self.is_directed = nx.is_directed(self.G)
        self.num_node = self.G.number_of_nodes()
        self.num_edge = self.G.number_of_edges()
        self.edges = [[node2id[e[0]], node2id[e[1]]] for e in self.G.edges()]

        id2node = dict(zip(node2id.values(), node2id.keys()))

        self.num_neigh = np.asarray([len(list(self.G.neighbors(id2node[i]))) for i in range(self.num_node)])
        self.neighbors = [[node2id[v] for v in self.G.neighbors(id2node[i])] for i in range(self.num_node)]
        s = time.time()
        self.alias_nodes = {}
        self.node_weight = {}
        for i in range(self.num_node):
            unnormalized_probs = [self.G[id2node[i]][nbr].get("weight", 1.0) for nbr in self.G.neighbors(id2node[i])]
            norm_const = sum(unnormalized_probs)
            normalized_probs = [float(u_prob) / norm_const for u_prob in unnormalized_probs]
            self.alias_nodes[i] = alias_setup(normalized_probs)
            self.node_weight[i] = dict(
                zip(
                    [node2id[nbr] for nbr in self.G.neighbors(id2node[i])],
                    unnormalized_probs,
                )
            )

        t = time.time()
        print("alias_nodes", t - s)

        # run netsmf algorithm with multiprocessing and apply randomized svd
        print("number of sample edges ", self.num_round * self.num_edge * self.window_size)
        print("random walk start...")
        t0 = time.time()
        results = []
        pool = Pool(processes=self.worker)
        for i in range(self.worker):
            results.append(pool.apply_async(func=self._random_walk_matrix, args=(i,)))
        pool.close()
        pool.join()
        print("random walk time", time.time() - t0)

        matrix = sp.csr_matrix((self.num_node, self.num_node))
        A = sp.csr_matrix(nx.adjacency_matrix(self.G))
        degree = sp.diags(np.array(A.sum(axis=0))[0], format="csr")
        degree_inv = degree.power(-1)

        t1 = time.time()
        for res in results:
            matrix += res.get()
        t2 = time.time()
        print("construct random walk matrix time", time.time() - t1)

        L = sp.csgraph.laplacian(matrix, normed=False, return_diag=False)
        M = degree_inv.dot(degree - L).dot(degree_inv)
        M = M * A.sum() / self.negative
        M.data[M.data <= 1] = 1
        M.data = np.log(M.data)
        M.eliminate_zeros()
        print("number of nzz", M.nnz)
        print("construct matrix sparsifier time", time.time() - t2)

        embeddings = self._get_embedding_rand(M)

        if return_dict:
            features_matrix = dict()
            for vid, node in enumerate(self.G.nodes()):
                features_matrix[node] = embeddings[vid]
        else:
            features_matrix = np.zeros((graph.num_nodes, embeddings.shape[1]))
            nx_nodes = self.G.nodes()
            features_matrix[nx_nodes] = embeddings[np.arange(graph.num_nodes)]
        return features_matrix

    def _get_embedding_rand(self, matrix):
        # Sparse randomized tSVD for fast embedding
        t1 = time.time()
        l = matrix.shape[0]  # noqa E741
        smat = sp.csc_matrix(matrix)
        print("svd sparse", smat.data.shape[0] * 1.0 / l ** 2)
        U, Sigma, VT = randomized_svd(smat, n_components=self.dimension, n_iter=5, random_state=None)
        U = U * np.sqrt(Sigma)
        U = preprocessing.normalize(U, "l2")
        print("sparsesvd time", time.time() - t1)
        return U

    def _path_sampling(self, u, v, r):
        # sample a r-length path from edge(u, v) and return path end node
        k = np.random.randint(r) + 1
        zp, rand_u, rand_v = 2.0 / self.node_weight[u][v], k - 1, r - k
        for i in range(rand_u):
            new_u = self.neighbors[u][alias_draw(self.alias_nodes[u][0], self.alias_nodes[u][1])]
            zp += 2.0 / self.node_weight[u][new_u]
            u = new_u
        for j in range(rand_v):
            new_v = self.neighbors[v][alias_draw(self.alias_nodes[v][0], self.alias_nodes[v][1])]
            zp += 2.0 / self.node_weight[v][new_v]
            v = new_v
        return u, v, zp

    def _random_walk_matrix(self, pid):
        # construct matrix based on random walk
        np.random.seed(pid)
        matrix = sp.lil_matrix((self.num_node, self.num_node))
        for i in tqdm(range(self.num_edge * self.num_round // self.worker)):
            u, v = self.edges[i % self.num_edge]
            if not self.is_directed and np.random.rand() > 0.5:
                v, u = u, v
            for r in range(1, self.window_size + 1):
                u_, v_, zp = self._path_sampling(u, v, r)
                matrix[u_, v_] += 2 * r / self.window_size / self.num_round / zp
        return matrix.tocsr()



import os
import os.path as osp
import numpy as np
from collections import defaultdict

import torch

from cogdl.data import Graph, Dataset
from cogdl.utils import download_url


class GCCDataset(Dataset):
    url = "https://github.com/cenyk1230/gcc-data/raw/master"

    def __init__(self, root, name):
        self.name = name
        super(GCCDataset, self).__init__(root)

        name1 = name.split("_")[0]
        name2 = name.split("_")[1]
        edge_index_1, dict_1, self.node2id_1 = self.preprocess(root, name1)
        edge_index_2, dict_2, self.node2id_2 = self.preprocess(root, name2)
        self.data = [
            Graph(x=None, edge_index=edge_index_1, name_dict=dict_1),
            Graph(x=None, edge_index=edge_index_2, name_dict=dict_2),
        ]
        self.transform = None

    @property
    def raw_file_names(self):

        names = [
            self.name.split("_")[0] + ".dict",
            self.name.split("_")[0] + ".graph",
            self.name.split("_")[1] + ".dict",
            self.name.split("_")[1] + ".graph",
        ]
        return names

    @property
    def processed_file_names(self):
        return []

    def get(self, idx):
        assert idx == 0
        return self.data

    def download(self):
        for name in self.raw_file_names:
            download_url("{}/{}/{}".format(self.url, self.name.lower(), name), self.raw_dir)

    def preprocess(self, root, name):
        dict_path = os.path.join(root, "raw/" + name + ".dict")
        graph_path = os.path.join(root, "raw/" + name + ".graph")

        with open(graph_path, "r") as f:
            edge_list = []
            node2id = defaultdict(int)
            f.readline()
            for line in f:
                x, y, t = list(map(int, line.strip().split()))
                # Reindex
                if x not in node2id:
                    node2id[x] = len(node2id)
                if y not in node2id:
                    node2id[y] = len(node2id)
                # repeat t times
                for _ in range(t):
                    # to undirected
                    edge_list.append([node2id[x], node2id[y]])
                    edge_list.append([node2id[y], node2id[x]])
        edge_list = torch.LongTensor(edge_list).t()

        name_dict = dict()
        with open(dict_path) as f:
            for line in f:
                name, str_x = line.split("\t")
                x = int(str_x)
                if x not in node2id:
                    node2id[x] = len(node2id)
                name_dict[name] = node2id[x]

        return (edge_list[0], edge_list[1]), name_dict, node2id

    def __repr__(self):
        return "{}()".format(self.name)


class Edgelist(Dataset):
    url = "https://github.com/cenyk1230/gcc-data/raw/master"

    def __init__(self, root, name):
        self.name = name
        super(Edgelist, self).__init__(root)
        self.data = torch.load(self.processed_paths[0])

    @property
    def raw_file_names(self):
        names = ["edgelist.txt", "nodelabel.txt"]
        return names

    @property
    def processed_file_names(self):
        return ["data.pt"]

    @property
    def num_classes(self):
        return self.data.y.shape[1]

    def download(self):
        for name in self.raw_file_names:
            download_url("{}/{}/{}".format(self.url, self.name.lower(), name), self.raw_dir)

    def get(self, idx):
        assert idx == 0
        return self.data

    def process(self):
        edge_list_path = osp.join(self.raw_dir, "edgelist.txt")
        node_label_path = osp.join(self.raw_dir, "nodelabel.txt")

        with open(edge_list_path) as f:
            edge_list = []
            node2id = defaultdict(int)
            for line in f:
                x, y = list(map(int, line.split()))
                # Reindex
                if x not in node2id:
                    node2id[x] = len(node2id)
                if y not in node2id:
                    node2id[y] = len(node2id)
                edge_list.append([node2id[x], node2id[y]])
                edge_list.append([node2id[y], node2id[x]])

        num_nodes = len(node2id)
        with open(node_label_path) as f:
            nodes = []
            labels = []
            label2id = defaultdict(int)
            for line in f:
                x, label = list(map(int, line.split()))
                if label not in label2id:
                    label2id[label] = len(label2id)
                nodes.append(node2id[x])
                if "hindex" in self.name:
                    labels.append(label)
                else:
                    labels.append(label2id[label])
            if "hindex" in self.name:
                median = np.median(labels)
                labels = [int(label > median) for label in labels]
        assert num_nodes == len(set(nodes))
        y = torch.zeros(num_nodes, len(label2id))
        y[nodes, labels] = 1

        data = Graph(edge_index=torch.LongTensor(edge_list).t(), x=None, y=y)

        torch.save(data, self.processed_paths[0])


class KDD_ICDM_GCCDataset(GCCDataset):
    def __init__(self, data_path="data"):
        dataset = "kdd_icdm"
        path = osp.join(data_path, dataset)
        super(KDD_ICDM_GCCDataset, self).__init__(path, dataset)


class SIGIR_CIKM_GCCDataset(GCCDataset):
    def __init__(self, data_path="data"):
        dataset = "sigir_cikm"
        path = osp.join(data_path, dataset)
        super(SIGIR_CIKM_GCCDataset, self).__init__(path, dataset)


class SIGMOD_ICDE_GCCDataset(GCCDataset):
    def __init__(self, data_path="data"):
        dataset = "sigmod_icde"
        path = osp.join(data_path, dataset)
        super(SIGMOD_ICDE_GCCDataset, self).__init__(path, dataset)


class USAAirportDataset(Edgelist):
    def __init__(self, data_path="data"):
        dataset = "usa-airport"
        path = osp.join(data_path, dataset)
        super(USAAirportDataset, self).__init__(path, dataset)



import torch.nn as nn
import torch.nn.functional as F
from cogdl.utils import get_activation
from cogdl.layers import ResGNNLayer, GENConv

from .. import BaseModel


class DeeperGCN(BaseModel):
    """Implementation of DeeperGCN in paper `"DeeperGCN: All You Need to Train Deeper GCNs" <https://arxiv.org/abs/2006.07739>`_

    Args:
        in_feat (int): the dimension of input features
        hidden_size (int): the dimension of hidden representation
        out_feat (int): the dimension of output features
        num_layers (int): the number of layers
        activation (str, optional): activation function. Defaults to "relu".
        dropout (float, optional): dropout rate. Defaults to 0.0.
        aggr (str, optional): aggregation function. Defaults to "max".
        beta (float, optional): a coefficient for aggregation function. Defaults to 1.0.
        p (float, optional): a coefficient for aggregation function. Defaults to 1.0.
        learn_beta (bool, optional): whether beta is learnable. Defaults to False.
        learn_p (bool, optional): whether p is learnable. Defaults to False.
        learn_msg_scale (bool, optional): whether message scale is learnable. Defaults to True.
        use_msg_norm (bool, optional): use message norm or not. Defaults to False.
        edge_attr_size (int, optional): the dimension of edge features. Defaults to None.
    """

    @staticmethod
    def add_args(parser):
        # fmt: off
        parser.add_argument("--num-features", type=int)
        parser.add_argument("--num-classes", type=int)
        parser.add_argument("--num-layers", type=int, default=14)
        parser.add_argument("--hidden-size", type=int, default=128)
        parser.add_argument("--dropout", type=float, default=0.5)
        parser.add_argument("--activation", type=str, default="relu")
        parser.add_argument("--aggr", type=str, default="softmax_sg")
        parser.add_argument("--beta", type=float, default=1.0)
        parser.add_argument("--p", type=float, default=1.0)
        parser.add_argument("--learn-beta", action="store_true")
        parser.add_argument("--learn-p", action="store_true")
        parser.add_argument("--learn-msg-scale", action="store_true")
        parser.add_argument("--use-msg-norm", action="store_true")
        # fmt: on

    @classmethod
    def build_model_from_args(cls, args):
        return cls(
            in_feat=args.num_features,
            hidden_size=args.hidden_size,
            out_feat=args.num_classes,
            num_layers=args.num_layers,
            activation=args.activation,
            dropout=args.dropout,
            aggr=args.aggr,
            beta=args.beta,
            p=args.p,
            learn_beta=args.learn_beta,
            learn_p=args.learn_p,
            learn_msg_scale=args.learn_msg_scale,
            use_msg_norm=args.use_msg_norm,
            edge_attr_size=args.edge_attr_size,
        )

    def __init__(
        self,
        in_feat,
        hidden_size,
        out_feat,
        num_layers,
        activation="relu",
        dropout=0.0,
        aggr="max",
        beta=1.0,
        p=1.0,
        learn_beta=False,
        learn_p=False,
        learn_msg_scale=True,
        use_msg_norm=False,
        edge_attr_size=None,
    ):
        super(DeeperGCN, self).__init__()
        self.dropout = dropout
        self.feat_encoder = nn.Linear(in_feat, hidden_size)

        self.layers = nn.ModuleList()
        for i in range(num_layers - 1):
            self.layers.append(
                ResGNNLayer(
                    conv=GENConv(
                        in_feats=hidden_size,
                        out_feats=hidden_size,
                        aggr=aggr,
                        beta=beta,
                        p=p,
                        learn_beta=learn_beta,
                        learn_p=learn_p,
                        use_msg_norm=use_msg_norm,
                        learn_msg_scale=learn_msg_scale,
                        edge_attr_size=edge_attr_size,
                    ),
                    in_channels=hidden_size,
                    activation=activation,
                    dropout=dropout,
                    checkpoint_grad=False,
                )
            )
        self.norm = nn.BatchNorm1d(hidden_size, affine=True)
        self.activation = get_activation(activation)
        self.fc = nn.Linear(hidden_size, out_feat)

    def forward(self, graph):
        x = graph.x
        h = self.feat_encoder(x)
        for layer in self.layers:
            h = layer(graph, h)
        h = self.activation(self.norm(h))
        h = F.dropout(h, p=self.dropout, training=self.training)
        h = self.fc(h)
        return h

    def predict(self, graph):
        return self.forward(graph)



import os
import numpy as np
import torch
from torch.utils.cpp_extension import load

path = os.path.join(os.path.dirname(__file__))

# SPMM

try:
    spmm_max = load(
        name="scatter_max",
        sources=[os.path.join(path, "scatter_max/scatter_max.cc"), os.path.join(path, "scatter_max/scatter_max.cu")],
        verbose=True,
    )

    def scatter_max(rowptr, colind, feat):
        return ScatterMaxFunction.apply(rowptr, colind, feat)


except Exception:
    spmm_max = None


class ScatterMaxFunction(torch.autograd.Function):
    @staticmethod
    def forward(ctx, rowptr, colind, feat):
        out, max_id = spmm_max.scatter_max_fp(rowptr, colind, feat)
        ctx.save_for_backward(max_id)
        return out

    @staticmethod
    def backward(ctx, grad):
        grad = grad.contiguous()
        max_id = ctx.saved_tensors[0]
        out = spmm_max.scatter_max_bp(grad, max_id)
        return None, None, out



import numpy as np
import torch.nn as nn
import torch.nn.functional as F

from cogdl.layers import SELayer, GCNLayer

from .. import BaseModel


class DrGCN(BaseModel):
    @staticmethod
    def add_args(parser):
        """Add model-specific arguments to the parser."""
        # fmt: off
        parser.add_argument("--num-features", type=int)
        parser.add_argument("--num-classes", type=int)
        parser.add_argument("--hidden-size", type=int, default=16)
        parser.add_argument("--num-layers", type=int, default=2)
        parser.add_argument("--dropout", type=float, default=0.5)
        parser.add_argument("--norm", type=str, default=None)
        parser.add_argument("--activation", type=str, default="relu")
        # fmt: on

    @classmethod
    def build_model_from_args(cls, args):
        return cls(
            args.num_features,
            args.num_classes,
            args.hidden_size,
            args.num_layers,
            args.dropout,
            args.norm,
            args.activation,
        )

    def __init__(self, num_features, num_classes, hidden_size, num_layers, dropout, norm=None, activation="relu"):
        super(DrGCN, self).__init__()

        self.num_features = num_features
        self.num_classes = num_classes
        self.hidden_size = hidden_size
        self.num_layers = num_layers
        self.dropout = dropout
        shapes = [num_features] + [hidden_size] * (num_layers - 1) + [num_classes]
        self.convs = nn.ModuleList(
            [
                GCNLayer(shapes[layer], shapes[layer + 1], activation=activation, norm=norm)
                for layer in range(num_layers - 1)
            ]
        )
        self.convs.append(GCNLayer(shapes[-2], shapes[-1]))
        self.ses = nn.ModuleList(
            [SELayer(shapes[layer], se_channels=int(np.sqrt(shapes[layer]))) for layer in range(num_layers)]
        )

    def forward(self, graph):
        graph.sym_norm()
        x = graph.x
        x = self.ses[0](x)
        for se, conv in zip(self.ses[1:], self.convs[:-1]):
            x = conv(graph, x)
            x = se(x)
            x = F.dropout(x, p=self.dropout, training=self.training)
        x = self.convs[-1](graph, x)
        return x

    def predict(self, graph):
        return self.forward(graph)



import torch
import torch.nn as nn

from .gat_layer import GATLayer


class AttentionLayer(nn.Module):
    def __init__(self, num_features):
        super(AttentionLayer, self).__init__()
        self.linear = nn.Linear(num_features, 1)

    def forward(self, x):
        att = self.linear(x).view(-1, 1, x.shape[1])
        return torch.matmul(att, x).squeeze(1)


class HANLayer(nn.Module):
    def __init__(self, num_edge, w_in, w_out):
        super(HANLayer, self).__init__()
        self.gat_layer = nn.ModuleList()
        for _ in range(num_edge):
            self.gat_layer.append(GATLayer(w_in, w_out // 8, 8))
        self.att_layer = AttentionLayer(w_out)

    def forward(self, graph, x):
        adj = graph.adj
        output = []
        with graph.local_graph():
            for i, edge in enumerate(adj):
                graph.edge_index = edge[0]
                output.append(self.gat_layer[i](graph, x))
        output = torch.stack(output, dim=1)

        return self.att_layer(output)



import math
import torch
import torch.nn.functional as F


try:
    from actnn.ops import quantize_activation, dequantize_activation
    from actnn.conf import config
    from actnn.utils import empty_cache
except Exception:
    pass


class linear(torch.autograd.Function):
    @staticmethod
    def forward(ctx, input, weight, bias=None, scheme=None, rp_ratio=2):
        if rp_ratio > 1:
            D = input.shape[1]
            rmat = (torch.bernoulli(torch.ones((D, D // rp_ratio)).to(input.device) * 0.5) * 2.0 - 1) * math.sqrt(
                1.0 / (D // rp_ratio)
            )
            input_rp = torch.mm(input, rmat)
            quantized = quantize_activation(input_rp, scheme)
        else:
            quantized = quantize_activation(input, scheme)

        empty_cache(config.empty_cache_threshold)

        ctx.scheme = scheme
        if rp_ratio > 1:
            ctx.saved = quantized, weight, bias, rmat
            ctx.other_args = input_rp.shape
        else:
            ctx.saved = quantized, weight, bias
            ctx.other_args = input.shape
        return F.linear(input, weight, bias)

    @staticmethod
    def backward(ctx, grad_output):
        if ctx.scheme:
            ctx.scheme.set_scale(grad_output)

        q_input_shape = ctx.other_args
        if len(ctx.saved) == 4:
            quantized, weight, bias, rmat = ctx.saved
            input_rp = dequantize_activation(quantized, q_input_shape)
            input = torch.mm(input_rp, rmat.t())
            del quantized, ctx.saved, input_rp
        else:
            quantized, weight, bias = ctx.saved
            input = dequantize_activation(quantized, q_input_shape)
            del quantized, ctx.saved

        empty_cache(config.empty_cache_threshold)

        C_in = input.shape[-1]
        C_out = grad_output.shape[-1]

        grad_output_flatten = grad_output.view(-1, C_out)
        input_flatten = input.view(-1, C_in)
        grad_input = grad_output_flatten.mm(weight)
        grad_weight = grad_output_flatten.t().mm(input_flatten)

        if bias is not None:
            grad_bias = grad_output_flatten.sum(0)
        else:
            grad_bias = None

        if ctx.scheme:
            ctx.scheme.if_allocate_perlayer()
        return grad_input, grad_weight, grad_bias, None, None



import numpy as np
import networkx as nx
import scipy.sparse as sp
from sklearn import preprocessing
from .. import BaseModel


class HOPE(BaseModel):
    r"""The HOPE model from the `"Grarep: Asymmetric transitivity preserving graph embedding"
    <http://dl.acm.org/citation.cfm?doid=2939672.2939751>`_ paper.

    Args:
        hidden_size (int) : The dimension of node representation.
        beta (float) : Parameter in katz decomposition.
    """

    @staticmethod
    def add_args(parser):
        """Add model-specific arguments to the parser."""
        # fmt: off
        parser.add_argument("--beta", type=float, default=0.01,
                            help="Parameter of katz for HOPE. Default is 0.01")
        parser.add_argument("--hidden-size", type=int, default=128)
        # fmt: on

    @classmethod
    def build_model_from_args(cls, args):
        return cls(args.hidden_size, args.beta)

    def __init__(self, dimension, beta):
        super(HOPE, self).__init__()
        self.dimension = dimension
        self.beta = beta

    def forward(self, graph, return_dict=False):
        r"""The author claim that Katz has superior performance in related tasks
        S_katz = (M_g)^-1 * M_l = (I - beta*A)^-1 * beta*A = (I - beta*A)^-1 * (I - (I -beta*A))
        = (I - beta*A)^-1 - I
        """
        nx_g = graph.to_networkx()
        adj = nx.adjacency_matrix(nx_g).todense()
        n = adj.shape[0]
        katz_matrix = np.asarray((np.eye(n) - self.beta * np.mat(adj)).I - np.eye(n))
        embeddings = self._get_embedding(katz_matrix, self.dimension)

        if return_dict:
            features_matrix = dict()
            for vid, node in enumerate(nx_g.nodes()):
                features_matrix[node] = embeddings[vid]
        else:
            features_matrix = np.zeros((graph.num_nodes, embeddings.shape[1]))
            nx_nodes = nx_g.nodes()
            features_matrix[nx_nodes] = embeddings[np.arange(graph.num_nodes)]
        return features_matrix

    def _get_embedding(self, matrix, dimension):
        # get embedding from svd and process normalization for ut and vt
        ut, s, vt = sp.linalg.svds(matrix, int(dimension / 2))
        emb_matrix_1, emb_matrix_2 = ut, vt.transpose()

        emb_matrix_1 = emb_matrix_1 * np.sqrt(s)
        emb_matrix_2 = emb_matrix_2 * np.sqrt(s)
        emb_matrix_1 = preprocessing.normalize(emb_matrix_1, "l2")
        emb_matrix_2 = preprocessing.normalize(emb_matrix_2, "l2")
        features = np.hstack((emb_matrix_1, emb_matrix_2))
        return features



import numpy as np
import tvm
from tvm import te
from tvm.topi.util import get_const_tuple

from ..util import util_partition_adj_coo_2d
from ..op import vanilla_sddmm, schedule_vanilla_sddmm_x86, \
    schedule_vanilla_sddmm_cuda_tree_reduce, schedule_vanilla_sddmm_cuda_single_thread_reduce


class SDDMMbase():
    """The base class for SDDMM-like computation kernels."""
    def __init__(self, adj_scipy, num_row_partitions=1, num_col_partitions=1):
        """Doing 2D graph partitioning during init.

        Parameters
        ----------
        adj_scipy : scipy.sparse.coo_matrix or scipy.sparse.csr_matrix
            The input scipy sparse matrix

        num_row_partitions : int
            Number of partitions along the row dimension

        num_col_partitions : int
            Number of partitions along the col dimension
        """
        # Use coo format in SDDMM-like kernels
        if adj_scipy.format != 'coo':
            adj_scipy_coo = adj_scipy.tocoo()
        else:
            adj_scipy_coo = adj_scipy
        self._num_rows = adj_scipy_coo.shape[0]
        self._num_cols = adj_scipy_coo.shape[1]
        assert num_row_partitions >= 1, "num_row_partitions should be larger than or equal to 1"
        assert num_col_partitions >= 1, "num_col_partitions should be larger than or equal to 1"
        self._num_row_partitions = num_row_partitions
        self._num_col_partitions = num_col_partitions
        # To be updated in self.register
        self._target = None
        self._ctx = None
        self._compute_func = None
        self._schedule_func = None
        self._register()
        # 2D graph partitioning
        if self._num_row_partitions > 1 or self._num_col_partitions > 1:
            edge_id_list, adj_row_indices, adj_col_indices = self._preprocess_adj(adj_scipy_coo, \
                self._num_row_partitions, self._num_col_partitions)
            # This is smart; credit to Zihao
            self._edge_mapping = np.argsort(edge_id_list)
        else:
            adj_row_indices = adj_scipy_coo.row
            adj_col_indices = adj_scipy_coo.col
        self._adj_row_indices = adj_row_indices
        self._adj_col_indices = adj_col_indices
        self._adj_row_indices_placeholder = te.placeholder(shape=self._adj_row_indices.shape, \
            dtype=str(self._adj_row_indices.dtype), name='adj_row_indices_placeholder')
        self._adj_col_indices_placeholder = te.placeholder(shape=self._adj_col_indices.shape, \
            dtype=str(self._adj_col_indices.dtype), name='adj_col_indices_placeholder')
        self._adj_row_indices_tvm = tvm.nd.array(self._adj_row_indices, ctx=self._ctx)
        self._adj_col_indices_tvm = tvm.nd.array(self._adj_col_indices, ctx=self._ctx)
        # To be updated in self.build
        self._func = None
        # To be updated in self.run
        self.out_tvm = None

    def _preprocess_adj(self, adj_scipy_coo, num_row_partitions=1, num_col_partitions=1):
        return util_partition_adj_coo_2d(adj_scipy_coo, num_row_partitions, num_col_partitions)

    def _register(self):
        raise NotImplementedError("Please register target, ctx, compute_func, and schedule_func.")

    def build(self, input_placeholders, compute_args, schedule_args):
        """Build tvm func, update self._func inplace.

        Parameters
        ----------
        input_placeholders : list of te.placeholder
            The required input tvm placeholders other than adj (which has been passed in during self.init)

        compute_args : dict
            Arguments required for compute_func, e.g., num_feat_partitions

        schedule_args : dict
            Arguments required for schedule_func, e.g., num_cuda_blocks
        """
        out_placeholder = self._compute_func(*input_placeholders, self._adj_row_indices_placeholder, \
            self._adj_col_indices_placeholder, **compute_args)  # use ** to unpack dict into kwargs
        s = self._schedule_func(out_placeholder, **schedule_args)
        self._func = tvm.build(s, [*input_placeholders, self._adj_row_indices_placeholder, \
            self._adj_col_indices_placeholder, out_placeholder], target=self._target)
        self.out_tvm = tvm.nd.array(np.zeros(shape=get_const_tuple(out_placeholder.shape), \
            dtype=str(out_placeholder.dtype)), ctx=self._ctx)

    def lower_to_ir(self, input_placeholders, compute_args, schedule_args):
        """Return the IR. This can be useful for debug.

        Parameters
        ----------
        input_placeholders : list of te.placeholder
            The required input tvm placeholders other than adj (which has been passed in during self.init)

        compute_args : dict
            Arguments required for compute_func, e.g., num_feat_partitions

        schedule_args : dict
            Arguments required for schedule_func, e.g., num_cuda_blocks
        """
        out_placeholder = self._compute_func(*input_placeholders, self._adj_row_indices_placeholder, \
            self._adj_col_indices_placeholder, **compute_args)  # use ** to unpack dict into kwargs
        s = self._schedule_func(out_placeholder, **schedule_args)
        ir = tvm.lower(s, [*input_placeholders, self._adj_row_indices_placeholder, \
            self._adj_col_indices_placeholder, out_placeholder], simple_mode=True)
        return ir

    def run(self, input_tvm_ndarrays):
        """Run the built func with the given inputs input_tvm_ndarrays.

        Parameters
        ----------
        input_tvm_ndarrays : list of tvm.ndarray
            The required input tvm ndarrays other than adj

        Returns
        -------
        self.out_tvm: tvm.ndarray
            The output tvm ndarray
        """
        self._func(*input_tvm_ndarrays, self._adj_row_indices_tvm, self._adj_col_indices_tvm, self.out_tvm)
        return self.out_tvm

    def measure_average_time(self, input_tvm_ndarrays, num_runs):
        """Measure the run time of the built module using tvm time_evaluator.

        Parameters
        ----------
        input_tvm_ndarrays : list of tvm.ndarray
            The required input tvm ndarrays other than adj

        int : num_runs
            The number of runs

        Returns
        -------
        tcost: float32
            The average run time measured in seconds
        """
        timer = self._func.time_evaluator(self._func.entry_name, ctx=self._ctx, number=num_runs)
        tcost = timer(*input_tvm_ndarrays, self._adj_row_indices_tvm,
            self._adj_col_indices_tvm, self.out_tvm).mean
        return tcost

    @property
    def edge_mapping(self):
        assert self._target != 'cuda', "no graph partitioning on cuda, no edge_mapping"
        assert self._num_row_partitions > 1 or self._num_col_partitions > 1, \
            "no graph partitioning, no edge_mapping"
        return self._edge_mapping

    @property
    def ctx(self):
        return self._ctx

    @property
    def target(self):
        return self._target


class VanillaSDDMMx86(SDDMMbase):
    def __init__(self, adj_scipy, num_row_partitions=1, num_col_partitions=1):
        super(VanillaSDDMMx86, self).__init__(adj_scipy, num_row_partitions, num_col_partitions)

    def _register(self):
        self._target = 'llvm'
        self._ctx = tvm.cpu(0)
        self._compute_func = vanilla_sddmm
        self._schedule_func = schedule_vanilla_sddmm_x86


class VanillaSDDMMcuda(SDDMMbase):
    def __init__(self, adj_scipy):
        super(VanillaSDDMMcuda, self).__init__(adj_scipy, num_row_partitions=1, num_col_partitions=1)

    def _register(self):
        self._target = 'cuda'
        self._ctx = tvm.gpu(0)
        self._compute_func = vanilla_sddmm
        self._schedule_func = schedule_vanilla_sddmm_cuda_tree_reduce



import copy

import torch
import numpy as np

from cogdl.wrappers.data_wrapper import DataWrapper

from .node_classification_mw import NodeClfModelWrapper


class CorrectSmoothModelWrapper(NodeClfModelWrapper):
    @staticmethod
    def add_args(parser):
        pass

    def __init__(self, model, optimizer_cfg):
        super(CorrectSmoothModelWrapper, self).__init__(model, optimizer_cfg)
        self.model = model
        self.optimizer_cfg = optimizer_cfg

    def val_step(self, subgraph):
        graph = subgraph
        pred = self.model(graph)
        pred = self.model.postprocess(graph, pred)
        y = graph.y
        val_mask = graph.val_mask
        loss = self.default_loss_fn(pred[val_mask], y[val_mask])

        metric = self.evaluate(pred[val_mask], graph.y[val_mask], metric="auto")

        self.note("val_loss", loss.item())
        self.note("val_metric", metric)

    def test_step(self, batch):
        graph = batch
        pred = self.model(graph)
        pred = self.model.postprocess(graph, pred)
        test_mask = batch.test_mask
        loss = self.default_loss_fn(pred[test_mask], batch.y[test_mask])

        metric = self.evaluate(pred[test_mask], batch.y[test_mask], metric="auto")

        self.note("test_loss", loss.item())
        self.note("test_metric", metric)



import functools
import random

import networkx as nx
import numpy as np
import torch
import torch.nn as nn

from cogdl.utils import split_dataset_general

from .. import BaseModel


class PatchySAN(BaseModel):
    r"""The Patchy-SAN model from the `"Learning Convolutional Neural Networks for Graphs"
    <https://arxiv.org/abs/1605.05273>`_ paper.

    Args:
        batch_size (int) : The batch size of training.
        sample (int) : Number of chosen vertexes.
        stride (int) : Node selection stride.
        neighbor (int) : The number of neighbor for each node.
        iteration (int) : The number of training iteration.
    """

    @staticmethod
    def add_args(parser):
        parser.add_argument("--num-sample", default=30, type=int, help="Number of chosen vertexes")
        # parser.add_argument("--stride", default=1, type=int, help="Stride of chosen vertexes")
        parser.add_argument("--num-neighbor", default=10, type=int, help="Number of neighbor in constructing features")
        parser.add_argument("--iteration", default=5, type=int, help="Number of iteration")

    @classmethod
    def build_model_from_args(cls, args):
        return cls(
            args.num_features,
            args.num_classes,
            args.num_sample,
            # args.stride,
            args.num_neighbor,
            args.iteration,
        )

    @classmethod
    def split_dataset(cls, dataset, args):
        # process each graph and add it into Data() as attribute x
        for i, data in enumerate(dataset):
            new_feature = get_single_feature(
                dataset[i], args.num_features, args.num_classes, args.num_sample, args.num_neighbor, args.stride
            )
            dataset[i].x = torch.from_numpy(new_feature)

        return split_dataset_general(dataset, args)

    # def __init__(self, batch_size, num_features, num_classes, num_sample, stride, num_neighbor, iteration):
    def __init__(self, num_features, num_classes, num_sample, num_neighbor, iteration):
        super(PatchySAN, self).__init__()

        self.num_features = num_features
        self.num_classes = num_classes
        self.num_sample = num_sample
        self.num_neighbor = num_neighbor
        self.iteration = iteration

        # self.build_model(self.num_features, self.num_sample, self.num_neighbor, self.num_classes)
        self.build_model(self.num_features, self.num_sample, self.num_neighbor, self.num_classes)

    def build_model(self, num_channel, num_sample, num_neighbor, num_class):
        rep1, stride1 = 4, 4
        num_filter1, num_filter2 = 16, 8
        self.conv1 = nn.Conv1d(num_channel, num_filter1, rep1, stride=stride1, groups=1)
        self.conv2 = nn.Conv1d(num_filter1, num_filter2, num_neighbor, stride=1, groups=1)

        num_lin = (int(num_sample * num_neighbor / stride1) - num_neighbor + 1) * num_filter2
        self.lin1 = torch.nn.Linear(num_lin, 128)
        self.lin2 = torch.nn.Linear(128, num_class)

        self.nn = nn.Sequential(
            self.conv1,
            nn.ReLU(),
            self.conv2,
            nn.ReLU(),
            nn.Flatten(),
            self.lin1,
            nn.ReLU(),
            nn.Dropout(0.2),
            self.lin2,
            nn.Softmax(),
        )

    def forward(self, batch):
        logits = self.nn(batch.x)
        return logits


def assemble_neighbor(G, node, num_neighbor, sorted_nodes):
    """assemble neighbors for node with BFS strategy"""
    neighbors_dict = dict()
    new_neighbors_dict = dict()
    neighbors_dict[node] = 0
    new_neighbors_dict[node] = 0
    # assemble over K neighbors with BFS strategy
    while len(neighbors_dict) < num_neighbor and len(new_neighbors_dict) > 0:
        temp_neighbor_dict = dict()
        for v, d in new_neighbors_dict.items():
            for new_v in G.neighbors(v):
                if new_v not in temp_neighbor_dict:
                    temp_neighbor_dict[new_v] = d + 1
        n = len(neighbors_dict)
        for v, d in temp_neighbor_dict.items():
            if v not in neighbors_dict:
                neighbors_dict[v] = d
        new_neighbors_dict = temp_neighbor_dict
        # break if the number of neighbors do not increase
        if n == len(neighbors_dict):
            break

    # add dummy disconnected nodes if number is not suffice
    while len(neighbors_dict) < num_neighbor:
        rand_node = sorted_nodes[random.randint(0, len(sorted_nodes) - 1)][0]
        if rand_node not in neighbors_dict:
            neighbors_dict[rand_node] = 10
    return neighbors_dict


def cmp(s1, s2):
    list1 = [int(l) for l in s1.strip().split(" ")]  # noqa E741
    list2 = [int(l) for l in s2.strip().split(" ")]  # noqa E741
    i = 0
    while i < len(list1) and i < len(list2):
        if list1[i] < list2[i]:
            return -1
        if list1[i] > list2[i]:
            return 1
        i += 1
    if i < len(list1):
        return 1
    elif i < len(list2):
        return -1
    else:
        return 0


def one_dim_wl(graph_list, init_labels, iteration=5):
    """1-dimension Wl method used for node normalization for all the subgraphs"""
    sorted_labels = sorted(list(set(init_labels.values())))
    label_dict = dict([(label, index) for index, label in enumerate(sorted_labels)])
    graph_label_list = []
    for t in range(iteration):
        new_label_dict = dict()
        # get label for each node in each graph
        if t == 0:
            # get label according to nodes' attribute labels
            for i in range(len(graph_list)):
                labels = dict()
                for id, v in enumerate(graph_list[i].nodes()):
                    labels[v] = str(init_labels[v])
                    new_label_dict[labels[v]] = 1
                graph_label_list.append(labels)
        else:
            # get label according to neighbors' labels
            for i in range(len(graph_list)):
                labels = dict()
                for id, v in enumerate(graph_list[i].nodes()):
                    neighbor_labels = [graph_label_list[i][v2] for v2 in graph_list[i].neighbors(v)]
                    sorted_labels = [str(l) for l in sorted(neighbor_labels)]  # noqa E741
                    # concentrate node label and its sorted neighbors' labels
                    new_label = str(graph_label_list[i][v]) + " " + " ".join(sorted_labels)
                    new_label_dict[new_label] = 1
                    labels[v] = new_label
                graph_label_list[i] = labels.copy()

        # sort new labels with dictionary order
        sorted_list = sorted(new_label_dict.keys(), key=functools.cmp_to_key(cmp), reverse=False)
        for new_label in sorted_list:
            # add new label to the labels_dict
            if new_label not in label_dict:
                label_dict[new_label] = len(label_dict)

        # relabel node labels with new label
        for i in range(len(graph_list)):
            for id, v in enumerate(graph_list[i].nodes()):
                graph_label_list[i][v] = label_dict[graph_label_list[i][v]]
    return graph_label_list


def node_selection_with_1d_wl(G, features, num_channel, num_sample, num_neighbor, stride):
    """construct features for cnn"""
    X = np.zeros((num_channel, num_sample, num_neighbor), dtype=np.float32)
    node2id = dict([(node, vid) for vid, node in enumerate(G.nodes())])
    betweenness = nx.betweenness_centrality(G)
    sorted_nodes = sorted(betweenness.items(), key=lambda d: d[1], reverse=False)
    # obtain normalized neighbors' features for each vertex
    i = 0
    j = 0
    root_list = []
    distance_list = []
    graph_list = []
    while j < num_sample:
        if i < len(sorted_nodes):
            neighbors_dict = assemble_neighbor(G, sorted_nodes[i][0], num_neighbor, sorted_nodes)
            # construct subgraph and sort neighbors with a labeling measurement like degree centrality
            sub_g = G.subgraph(neighbors_dict.keys())
            root_list.append(sorted_nodes[i][0])
            distance_list.append(neighbors_dict)
            graph_list.append(sub_g)
        else:
            # zero receptive field
            X[:, j, :] = np.zeros((num_channel, num_neighbor), dtype=np.float32)
        i += stride
        j += 1
    init_labels = dict([(v, features[id].argmax(axis=0)) for v, id in node2id.items()])
    graph_labels_list = one_dim_wl(graph_list, init_labels)

    # finally relabel based on 1d-wl and distance to root node
    for i in range(len(root_list)):
        # set root node the first position
        graph_labels_list[i][root_list[i]] = 0
        sorted_measurement = dict([(v, [measure, distance_list[i][v]]) for v, measure in graph_labels_list[i].items()])
        sorted_neighbor = sorted(sorted_measurement.items(), key=lambda d: d[1], reverse=False)[:num_neighbor]
        reorder_dict = dict(zip(sorted(sorted_measurement.keys()), range(len(sorted_measurement))))
        X[:, i, :] = features[[reorder_dict[v] for v, measure in sorted_neighbor]].T
    return X.reshape(num_channel, num_sample * num_neighbor)


def get_single_feature(data, num_features, num_classes, num_sample, num_neighbor, stride=1):
    """construct features"""
    data_list = [data]
    X = np.zeros((len(data_list), num_features, num_sample * num_neighbor), dtype=np.float32)
    for i in range(len(data_list)):
        edge_index, features = data_list[i].edge_index, data_list[i].x
        G = nx.Graph()
        row, col = edge_index[0].numpy(), edge_index[1].numpy()
        G.add_edges_from(list(zip(row, col)))
        # print("graph", i, "number of node", G.number_of_nodes(), "edge", G.number_of_edges())
        if G.number_of_nodes() > num_neighbor:
            X[i] = node_selection_with_1d_wl(G, features.cpu().numpy(), num_features, num_sample, num_neighbor, stride)
    X = X.astype(np.float32)
    return X



from cogdl.wrappers.tools.wrapper_utils import evaluate_clustering
from .. import EmbeddingModelWrapper


class AGCModelWrapper(EmbeddingModelWrapper):
    @staticmethod
    def add_args(parser):
        # fmt: off
        parser.add_argument("--num-clusters", type=int, default=7)
        parser.add_argument("--cluster-method", type=str, default="kmeans", help="option: kmeans or spectral")
        parser.add_argument("--evaluation", type=str, default="full", help="option: full or NMI")
        # fmt: on

    def __init__(self, model, optimizer_cfg, num_clusters, cluster_method="kmeans", evaluation="full", max_iter=5):
        super(AGCModelWrapper, self).__init__()
        self.model = model
        self.optimizer_cfg = optimizer_cfg
        self.num_clusters = num_clusters
        self.cluster_method = cluster_method
        self.full = evaluation == "full"

    def train_step(self, graph):
        emb = self.model.forward(graph)
        return emb

    def test_step(self, batch):
        features_matrix, graph = batch
        return evaluate_clustering(
            features_matrix, graph.y, self.cluster_method, self.num_clusters, graph.num_nodes, self.full
        )



from .. import DataWrapper


class HeterogeneousGNNDataWrapper(DataWrapper):
    def __init__(self, dataset):
        super(HeterogeneousGNNDataWrapper, self).__init__(dataset=dataset)

        self.dataset = dataset

    def train_wrapper(self):
        return self.dataset

    def val_wrapper(self):
        return self.dataset

    def test_wrapper(self):
        return self.dataset



import torch.nn as nn

from cogdl.utils import spmm


class SGCLayer(nn.Module):
    def __init__(self, in_features, out_features, order=3):
        super(SGCLayer, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.order = order
        self.W = nn.Linear(in_features, out_features)

    def forward(self, graph, x):
        output = self.W(x)
        for _ in range(self.order):
            output = spmm(graph, output)
        return output



import numpy as np
import scipy.sparse as sp
import torch
import torch.nn as nn

from .. import BaseModel
from .dgi import GCN
from cogdl.utils.ppr_utils import build_topk_ppr_matrix_from_data
from cogdl.data import Graph


def sparse_mx_to_torch_sparse_tensor(sparse_mx):
    """Convert a scipy sparse matrix to a torch sparse tensor."""
    sparse_mx = sparse_mx.tocoo().astype(np.float32)
    indices = torch.from_numpy(np.vstack((sparse_mx.row, sparse_mx.col)).astype(np.int64))
    values = torch.from_numpy(sparse_mx.data)
    shape = torch.Size(sparse_mx.shape)
    return torch.sparse.FloatTensor(indices, values, shape)


def compute_ppr(adj, index, alpha=0.4, epsilon=1e-4, k=8, norm="row"):
    return build_topk_ppr_matrix_from_data(adj, alpha, epsilon, index, k, norm).tocsr()


# Borrowed from https://github.com/PetarV-/DGI
class AvgReadout(nn.Module):
    def __init__(self):
        super(AvgReadout, self).__init__()

    def forward(self, seq, msk):
        dim = len(seq.shape) - 2
        if msk is None:
            return torch.mean(seq, dim)
        else:
            return torch.sum(seq * msk, dim) / torch.sum(msk)


# Borrowed from https://github.com/kavehhassani/mvgrl
class Discriminator(nn.Module):
    def __init__(self, n_h):
        super(Discriminator, self).__init__()
        self.f_k = nn.Bilinear(n_h, n_h, 1)

        for m in self.modules():
            self.weights_init(m)

    def weights_init(self, m):
        if isinstance(m, nn.Bilinear):
            torch.nn.init.xavier_uniform_(m.weight.data)
            if m.bias is not None:
                m.bias.data.fill_(0.0)

    def forward(self, c1, c2, h1, h2, h3, h4):
        c_x1 = torch.unsqueeze(c1, 0)
        c_x1 = c_x1.expand_as(h1).contiguous()
        c_x2 = torch.unsqueeze(c2, 0)
        c_x2 = c_x2.expand_as(h2).contiguous()

        # positive
        sc_1 = torch.squeeze(self.f_k(h2, c_x1), 1)
        sc_2 = torch.squeeze(self.f_k(h1, c_x2), 1)

        # negetive
        sc_3 = torch.squeeze(self.f_k(h4, c_x1), 1)
        sc_4 = torch.squeeze(self.f_k(h3, c_x2), 1)

        logits = torch.cat((sc_1, sc_2, sc_3, sc_4), 0)
        return logits


# Mainly borrowed from https://github.com/kavehhassani/mvgrl
class MVGRL(BaseModel):
    @staticmethod
    def add_args(parser):
        """Add model-specific arguments to the parser."""
        # fmt: off
        parser.add_argument("--hidden-size", type=int, default=512)
        parser.add_argument("--sample-size", type=int, default=2000)
        parser.add_argument("--batch-size", type=int, default=4)
        parser.add_argument("--alpha", type=float, default=0.2)
        # fmt: on

    @classmethod
    def build_model_from_args(cls, args):
        return cls(args.num_features, args.hidden_size, args.sample_size, args.batch_size, args.alpha, args.dataset)

    def __init__(self, in_feats, hidden_size, sample_size=2000, batch_size=4, alpha=0.2, dataset="cora"):
        super(MVGRL, self).__init__()
        self.sample_size = sample_size
        self.batch_size = batch_size
        self.hidden_size = hidden_size
        self.alpha = alpha
        self.sparse = True
        self.dataset_name = dataset

        self.gcn1 = GCN(in_feats, hidden_size, "prelu")
        self.gcn2 = GCN(in_feats, hidden_size, "prelu")
        self.read = AvgReadout()

        self.sigm = nn.Sigmoid()

        self.disc = Discriminator(hidden_size)
        self.loss_f = nn.BCEWithLogitsLoss()

        self.cache = None

    def _forward(self, adj, diff, seq1, seq2, msk):
        out_shape = list(seq1.shape[:-1]) + [self.hidden_size]

        seq1 = seq1.view(-1, seq1.shape[-1])
        seq2 = seq2.view(-1, seq2.shape[-1])

        h_1 = self.gcn1(adj, seq1, True)
        h_1 = h_1.view(out_shape)
        c_1 = self.read(h_1, msk)
        c_1 = self.sigm(c_1)

        h_2 = self.gcn2(diff, seq1, True)
        h_2 = h_2.view(out_shape)
        c_2 = self.read(h_2, msk)
        c_2 = self.sigm(c_2)

        h_3 = self.gcn1(adj, seq2, True)
        h_4 = self.gcn2(diff, seq2, True)
        h_3 = h_3.view(out_shape)
        h_4 = h_4.view(out_shape)

        ret = self.disc(c_1, c_2, h_1, h_2, h_3, h_4)

        return ret, h_1, h_2

    def augment(self, graph):
        num_nodes = graph.num_nodes
        adj = sp.coo_matrix(
            (graph.edge_weight.cpu().numpy(), (graph.edge_index[0].cpu().numpy(), graph.edge_index[1].cpu().numpy())),
            shape=(graph.num_nodes, graph.num_nodes),
        )
        diff = compute_ppr(adj.tocsr(), np.arange(num_nodes), self.alpha).tocoo()
        return adj, diff

    def preprocess(self, graph):
        print("MVGRL preprocessing...")
        graph.add_remaining_self_loops()
        graph.sym_norm()

        adj, diff = self.augment(graph)

        if self.cache is None:
            self.cache = dict()
        graphs = []
        for g in [adj, diff]:
            row = torch.from_numpy(g.row).long()
            col = torch.from_numpy(g.col).long()
            val = torch.from_numpy(g.data).float()
            edge_index = torch.stack([row, col])
            graphs.append(Graph(edge_index=edge_index, edge_weight=val))

        self.cache["diff"] = graphs[1]
        self.cache["adj"] = graphs[0]
        print("Preprocessing Done...")

    def forward(self, graph):
        if not self.training:
            return self.embed(graph)

        x = graph.x
        if self.cache is None or "diff" not in self.cache:
            self.preprocess(graph)
        diff, adj = self.cache["diff"], self.cache["adj"]

        self.sample_size = min(self.sample_size, graph.num_nodes - self.batch_size)
        idx = np.random.randint(0, graph.num_nodes - self.sample_size + 1, self.batch_size)
        logits = []
        for i in idx:
            ba = adj.subgraph(list(range(i, i + self.sample_size))).to(self.device)
            bd = diff.subgraph(list(range(i, i + self.sample_size))).to(self.device)
            bf = x[i : i + self.sample_size].to(self.device)
            idx = np.random.permutation(self.sample_size)
            shuf_fts = bf[idx, :].to(self.device)
            logit, _, _ = self._forward(ba, bd, bf, shuf_fts, None)
            logits.append(logit)

        return torch.stack(logits)

    def loss(self, data):
        if self.sample_size > data.num_nodes:
            self.sample_size = data.num_nodes
        if self.cache is None:
            self.device = next(self.gcn1.parameters()).device
            lbl_1 = torch.ones(self.batch_size, self.sample_size * 2)
            lbl_2 = torch.zeros(self.batch_size, self.sample_size * 2)
            lbl = torch.cat((lbl_1, lbl_2), 1)
            lbl = lbl.to(self.device)
            self.cache = {"labels": lbl}
        lbl = self.cache["labels"]
        logits = self.forward(data)
        loss = self.loss_f(logits, lbl)
        return loss

    def embed(self, data, msk=None):
        adj = self.cache["adj"].to(self.device)
        diff = self.cache["diff"].to(self.device)
        h_1 = self.gcn1(adj, data.x.to(self.device), True)
        h_2 = self.gcn2(diff, data.x.to(self.device), True)
        # c = self.read(h_1, msk)
        return (h_1 + h_2).detach()  # , c.detach()



from .base_data_wrapper import DataWrapper
import importlib


def register_data_wrapper(name):
    """
    New data wrapper types can be added to cogdl with the :func:`register_data_wrapper`
    function decorator.

    Args:
        name (str): the name of the data_wrapper
    """

    def register_data_wrapper_cls(cls):
        print("The `register_data_wrapper` API is deprecated!")
        return cls

    return register_data_wrapper_cls


def fetch_data_wrapper(name):
    if isinstance(name, type):
        return name
    if name in SUPPORTED_DW:
        path = ".".join(SUPPORTED_DW[name].split(".")[:-1])
        module = importlib.import_module(path)
    else:
        raise NotImplementedError(f"Failed to import {name} DataWrapper.")
    class_name = SUPPORTED_DW[name].split(".")[-1]
    return getattr(module, class_name)


SUPPORTED_DW = {
    "cluster_dw": "cogdl.wrappers.data_wrapper.node_classification.ClusterWrapper",
    "graphsage_dw": "cogdl.wrappers.data_wrapper.node_classification.GraphSAGEDataWrapper",
    "m3s_dw": "cogdl.wrappers.data_wrapper.node_classification.M3SDataWrapper",
    "network_embedding_dw": "cogdl.wrappers.data_wrapper.node_classification.NetworkEmbeddingDataWrapper",
    "node_classification_dw": "cogdl.wrappers.data_wrapper.node_classification.FullBatchNodeClfDataWrapper",
    "pprgo_dw": "cogdl.wrappers.data_wrapper.node_classification.PPRGoDataWrapper",
    "sagn_dw": "cogdl.wrappers.data_wrapper.node_classification.SAGNDataWrapper",
    "gcc_dw": "cogdl.wrappers.data_wrapper.pretraining.GCCDataWrapper",
    "embedding_link_prediction_dw": "cogdl.wrappers.data_wrapper.link_prediction.EmbeddingLinkPredictionDataWrapper",
    "gnn_kg_link_prediction_dw": "cogdl.wrappers.data_wrapper.link_prediction.GNNKGLinkPredictionDataWrapper",
    "gnn_link_prediction_dw": "cogdl.wrappers.data_wrapper.link_prediction.GNNLinkPredictionDataWrapper",
    "heterogeneous_embedding_dw": "cogdl.wrappers.data_wrapper.heterogeneous.HeterogeneousEmbeddingDataWrapper",
    "heterogeneous_gnn_dw": "cogdl.wrappers.data_wrapper.heterogeneous.HeterogeneousGNNDataWrapper",
    "multiplex_embedding_dw": "cogdl.wrappers.data_wrapper.heterogeneous.MultiplexEmbeddingDataWrapper",
    "graph_classification_dw": "cogdl.wrappers.data_wrapper.graph_classification.GraphClassificationDataWrapper",
    "graph_embedding_dw": "cogdl.wrappers.data_wrapper.graph_classification.GraphEmbeddingDataWrapper",
    "infograph_dw": "cogdl.wrappers.data_wrapper.graph_classification.InfoGraphDataWrapper",
    "patchy_san_dw": "cogdl.wrappers.data_wrapper.graph_classification.PATCHY_SAN_DataWrapper",
}



import torch

from .. import BaseModel
from cogdl.utils import spmm
from cogdl.layers import PPRGoLayer


class PPRGo(BaseModel):
    @staticmethod
    def add_args(parser):
        parser.add_argument("--hidden-size", type=int, default=32)
        parser.add_argument("--num-layers", type=int, default=2)
        parser.add_argument("--dropout", type=float, default=0.1)
        parser.add_argument("--activation", type=str, default="relu")
        parser.add_argument("--nprop-inference", type=int, default=2)
        parser.add_argument("--alpha", type=float, default=0.5)

    @classmethod
    def build_model_from_args(cls, args):
        return cls(
            in_feats=args.num_features,
            hidden_size=args.hidden_size,
            out_feats=args.num_classes,
            num_layers=args.num_layers,
            alpha=args.alpha,
            dropout=args.dropout,
            activation=args.activation,
            nprop=args.nprop_inference,
            norm=args.norm if hasattr(args, "norm") else "sym",
        )

    def __init__(
        self, in_feats, hidden_size, out_feats, num_layers, alpha, dropout, activation="relu", nprop=2, norm="sym"
    ):
        super(PPRGo, self).__init__()
        self.alpha = alpha
        self.norm = norm
        self.nprop = nprop
        self.fc = PPRGoLayer(in_feats, hidden_size, out_feats, num_layers, dropout, activation)

    def forward(self, x, targets, ppr_scores):
        h = self.fc(x)
        h = ppr_scores.unsqueeze(1) * h
        batch_size = targets[-1] + 1
        out = torch.zeros(batch_size, h.shape[1]).to(x.device).to(x.dtype)
        out = out.scatter_add_(dim=0, index=targets[:, None].repeat(1, h.shape[1]), src=h)
        return out

    def predict(self, graph, batch_size=10000):
        device = next(self.parameters()).device
        x = graph.x
        num_nodes = x.shape[0]
        pred_logits = []
        with torch.no_grad():
            for i in range(0, num_nodes, batch_size):
                batch_x = x[i : i + batch_size].to(device)
                batch_logits = self.fc(batch_x)
                pred_logits.append(batch_logits.cpu())
        pred_logits = torch.cat(pred_logits, dim=0)
        pred_logits = pred_logits.to(device)

        with graph.local_graph():
            if self.norm == "sym":
                graph.sym_norm()
            elif self.norm == "row":
                graph.row_norm()
            else:
                raise NotImplementedError
            edge_weight = graph.edge_weight * (1 - self.alpha)

            graph.edge_weight = edge_weight
            predictions = pred_logits
            for _ in range(self.nprop):
                predictions = spmm(graph, predictions) + self.alpha * pred_logits
        return predictions



from typing import Dict

import numpy as np

import torch
import torch.distributed as dist


def merge_batch_indexes(outputs: list):
    assert len(outputs) > 0
    keys = list(outputs[0].keys())

    results = dict()
    for key in keys:
        values = [x[key] for x in outputs]
        if key.endswith("loss"):
            results[key] = sum(values).item() / len(values)
        elif key.endswith("eval_index"):
            if len(values) > 1:
                val = torch.cat(values, dim=0)
                val = val.sum(0)
            else:
                val = values[0]
            fp = val[0]
            all_ = val.sum()

            prefix = key[: key.find("eval_index")]
            if val.shape[0] == 2:
                _key = prefix + "acc"
            else:
                _key = prefix + "f1"
            results[_key] = (fp / all_).item()
        else:
            results[key] = sum(values)
    return results


def bigger_than(x, y):
    return x >= y


def smaller_than(x, y):
    return x <= y


def evaluation_comp(monitor, compare="<"):
    if "loss" in monitor or compare == "<":
        return np.inf, smaller_than
    else:
        return 0, bigger_than


def save_model(model, path, epoch):
    print(f"Saving {epoch}-th model to {path} ...")
    torch.save(model.state_dict(), path)


def load_model(model, path):
    print(f"Loading model from {path} ...")
    model.load_state_dict(torch.load(path))
    return model


def ddp_after_epoch(*args):
    dist.barrier()


def ddp_end(*args):
    dist.barrier()
    dist.destroy_process_group()


class Printer(object):
    def __init__(self, print_fn, rank=0, world_size=1):
        self.printer = print_fn
        self.to_print = (world_size <= 1) or rank == 0 or rank == "cpu"

    def __call__(self, k_v: Dict):
        if self.to_print:
            assert "Epoch" in k_v
            out = f"Epoch: {k_v['Epoch']}"
            k_v.pop("Epoch")

            for k, v in k_v.items():
                if isinstance(v, float):
                    out += f", {k}: {v: .4f}"
                else:
                    out += f", {k}: {v}"
            self.printer(out)



from .vanilla_sddmm import vanilla_sddmm, schedule_vanilla_sddmm_x86, \
    schedule_vanilla_sddmm_cuda_tree_reduce, schedule_vanilla_sddmm_cuda_single_thread_reduce
from .vanilla_spmm import vanilla_spmm_csr_x86, schedule_vanilla_spmm_csr_x86, \
    vanilla_spmm_dds_x86, schedule_vanilla_spmm_dds_x86, \
    vanilla_spmm_csr_cuda, schedule_vanilla_spmm_csr_cuda



from typing import Optional

import torch
import torch.nn as nn
import torch.nn.functional as F

from .. import BaseModel
from .deepergcn import DeeperGCN
from .gat import GAT
from cogdl.layers.reversible_layer import RevGNNLayer
from cogdl.layers import GCNLayer, GATLayer, GENConv, ResGNNLayer
from cogdl.utils import get_activation, get_norm_layer, dropout_adj


def shared_dropout(x, dropout):
    m = torch.zeros_like(x).bernoulli_(1 - dropout)
    mask = m.requires_grad_(False) / (1 - dropout)
    return mask


class RevGCN(BaseModel):
    @staticmethod
    def add_args(parser):
        parser.add_argument("--group", type=int, default=2)
        parser.add_argument("--drop-edge-rate", type=float, default=0.0)
        parser.add_argument("--hidden-size", type=int, default=64)
        parser.add_argument("--dropout", type=float, default=0.5)
        parser.add_argument("--norm", type=str, default="batchnorm")
        parser.add_argument("--activation", type=str, default="relu")
        parser.add_argument("--num-layers", type=int, default=2)

    @classmethod
    def build_model_from_args(cls, args):
        return cls(
            args.num_features,
            args.num_classes,
            args.hidden_size,
            args.num_layers,
            args.dropout,
            args.drop_edge_rate,
            args.activation,
            args.norm,
            args.group,
        )

    def __init__(
        self,
        in_feats,
        out_feats,
        hidden_size,
        num_layers,
        dropout=0.5,
        drop_edge_rate=0.1,
        activation="relu",
        norm="batchnorm",
        group=2,
    ):
        super(RevGCN, self).__init__()
        self.dropout = dropout
        self.drop_edge_rate = drop_edge_rate
        self.num_layers = num_layers
        self.layers = nn.ModuleList()
        self.norm = get_norm_layer(norm, hidden_size)
        self.act = get_activation(activation)

        for i in range(num_layers):
            if i == 0:
                self.layers.append(
                    GCNLayer(
                        in_feats,
                        hidden_size,
                        residual=True,
                    )
                )
            elif i == num_layers - 1:
                self.layers.append(GCNLayer(hidden_size, out_feats, residual=True))
            else:
                conv = GCNLayer(
                    hidden_size // group,
                    hidden_size // group,
                )
                res_conv = ResGNNLayer(
                    conv,
                    hidden_size // group,
                    activation=activation,
                    norm=norm,
                    out_norm=norm,
                    out_channels=hidden_size // group,
                )
                self.layers.append(RevGNNLayer(res_conv, group))

    def forward(self, graph):
        graph.requires_grad = False
        edge_index, edge_weight = dropout_adj(
            graph.edge_index, drop_rate=self.drop_edge_rate, renorm=None, training=self.training
        )
        h = graph.x
        h = F.dropout(h, self.dropout, training=self.training)

        with graph.local_graph():
            graph.edge_index = edge_index
            graph.sym_norm()
            assert (graph.degrees() > 0).all()
            h = self.layers[0](graph, h)

            mask = shared_dropout(h, self.dropout)
            for i in range(1, len(self.layers) - 1):
                h = self.layers[i](graph, h, mask)
            h = self.norm(h)
            h = self.act(h)
            h = F.dropout(h, p=self.dropout, training=self.training)
            h = self.layers[-1](graph, h)

        return h


class RevGEN(BaseModel):
    @staticmethod
    def add_args(parser):
        DeeperGCN.add_args(parser)
        parser.add_argument("--group", type=int, default=2)
        parser.add_argument("--norm", type=str, default="batchnorm")
        parser.add_argument("--last-norm", type=str, default="batchnorm")
        parser.add_argument("--use-one-hot-emb", action="store_true")

    @classmethod
    def build_model_from_args(cls, args):
        return cls(
            args.num_features,
            args.hidden_size,
            args.num_classes,
            args.num_layers,
            args.group,
            args.activation,
            args.norm,
            args.last_norm,
            args.dropout,
            args.aggr,
            args.beta,
            args.p,
            args.learn_beta,
            args.learn_p,
            args.learn_msg_scale,
            args.use_msg_norm,
            edge_attr_size=args.edge_attr_size,
            one_hot_emb=args.use_one_hot_emb if hasattr(args, "use_one_hot_emb") else False,
        )

    def __init__(
        self,
        in_feats,
        hidden_size,
        out_feats,
        num_layers,
        group=2,
        activation="relu",
        norm="batchnorm",
        last_norm="batchnorm",
        dropout=0.0,
        aggr="softmax_sg",
        beta=1.0,
        p=1.0,
        learn_beta=False,
        learn_p=False,
        learn_msg_scale=True,
        use_msg_norm=False,
        edge_attr_size: Optional[list] = None,
        one_hot_emb: bool = False,
    ):
        super(RevGEN, self).__init__()
        self.input_fc = nn.Linear(in_feats, hidden_size)
        self.output_fc = nn.Linear(hidden_size, out_feats)
        self.layers = nn.ModuleList()

        for _ in range(num_layers):
            conv = GENConv(
                hidden_size // group,
                hidden_size // group,
                aggr,
                beta,
                p,
                learn_beta,
                learn_p,
                use_msg_norm,
                learn_msg_scale,
                residual=True,
                edge_attr_size=edge_attr_size,
            )
            res_conv = ResGNNLayer(conv, hidden_size // group, norm=norm, activation=activation, residual=False)
            self.layers.append(RevGNNLayer(res_conv, group))
        self.activation = get_activation(activation)
        self.norm = get_norm_layer(last_norm, hidden_size)
        self.dropout = dropout
        if one_hot_emb:
            self.one_hot_encoder = nn.Linear(in_feats // 2, in_feats // 2)
        self.use_one_hot_emb = one_hot_emb

    def forward(self, graph):
        graph.requires_grad = False
        x = graph.x
        if self.use_one_hot_emb:
            x = x.split(2, dim=-1)
            x[1] = self.one_hot_encoder(x[1])
            x = torch.cat((x[0], x[1]), dim=1)
        h = self.input_fc(x)

        mask = shared_dropout(h, self.dropout)

        for layer in self.layers:
            h = layer(graph, h, mask)

        h = self.activation(self.norm(h))
        h = F.dropout(h, p=self.dropout, training=self.training)
        h = self.output_fc(h)
        return h


class RevGAT(BaseModel):
    @staticmethod
    def add_args(parser):
        GAT.add_args(parser)
        parser.add_argument("--norm", type=str, default="batchnorm")
        parser.add_argument("--activation", type=str, default="relu")
        parser.add_argument("--group", type=int, default=2)

    @classmethod
    def build_model_from_args(cls, args):
        return cls(
            args.num_features,
            args.hidden_size,
            args.num_classes,
            args.num_layers,
            args.group,
            args.alpha,
            args.nhead,
            args.dropout,
            args.attn_drop,
            args.activation,
            args.norm,
        )

    def __init__(
        self,
        in_feats,
        hidden_size,
        out_feats,
        num_layers,
        group=2,
        alpha=0.2,
        nhead=1,
        dropout=0.5,
        attn_drop=0.5,
        activation="relu",
        norm="batchnorm",
    ):
        super(RevGAT, self).__init__()
        self.dropout = dropout
        self.num_layers = num_layers
        self.layers = nn.ModuleList()
        self.norm = get_norm_layer(norm, hidden_size * nhead)
        self.act = get_activation(activation)
        for i in range(num_layers):
            if i == 0:
                self.layers.append(
                    GATLayer(
                        in_feats,
                        hidden_size,
                        nhead,
                        alpha,
                        attn_drop,
                        residual=True,
                    )
                )
            elif i == num_layers - 1:
                self.layers.append(GATLayer(hidden_size * nhead, out_feats, 1, alpha, attn_drop, residual=True))
            else:
                conv = GATLayer(
                    hidden_size * nhead // group,
                    hidden_size // group,
                    nhead=nhead,
                    alpha=alpha,
                    attn_drop=attn_drop,
                )
                res_conv = ResGNNLayer(
                    conv,
                    hidden_size * nhead // group,
                    activation=activation,
                    norm=norm,
                    out_norm=norm,
                    out_channels=hidden_size * nhead // group,
                )
                self.layers.append(RevGNNLayer(res_conv, group))

    def forward(self, graph):
        graph.requires_grad = False
        h = graph.x
        h = F.dropout(h, self.dropout, training=self.training)
        h = self.layers[0](graph, h)

        mask = shared_dropout(h, self.dropout)
        for i in range(1, len(self.layers) - 1):
            h = self.layers[i](graph, h, mask)
            if torch.isnan(h).any():
                print(f"! NaN - h_{i}")
                input()

        h = self.norm(h)
        h = self.act(h)
        h = F.dropout(h, p=self.dropout, training=self.training)
        h = self.layers[-1](graph, h)
        return h



import math
import random

import torch
import torch.nn as nn
import torch.nn.functional as F

from .mlp import MLP
from cogdl.layers import GINLayer
from cogdl.utils import batch_mean_pooling, batch_sum_pooling, split_dataset_general
from .. import BaseModel


class Encoder(nn.Module):
    r"""Encoder stacked with GIN layers

    Parameters
    ----------
    in_feats : int
        Size of each input sample.
    hidden_feats : int
        Size of output embedding.
    num_layers : int, optional
        Number of GIN layers, default: ``3``.
    num_mlp_layers : int, optional
        Number of MLP layers for each GIN layer, default: ``2``.
    pooling : str, optional
        Aggragation type, default : ``sum``.

    """

    def __init__(self, in_feats, hidden_dim, num_layers=3, num_mlp_layers=2, pooling="sum"):
        super(Encoder, self).__init__()
        self.num_layers = num_layers
        self.gnn_layers = nn.ModuleList()
        self.bn_layers = nn.ModuleList()
        for i in range(num_layers):
            if i == 0:
                mlp = MLP(in_feats, hidden_dim, hidden_dim, num_mlp_layers, norm="batchnorm")
            else:
                mlp = MLP(hidden_dim, hidden_dim, hidden_dim, num_mlp_layers, norm="batchnorm")
            self.gnn_layers.append(GINLayer(mlp, eps=0, train_eps=True))
            self.bn_layers.append(nn.BatchNorm1d(hidden_dim))

        if pooling == "sum":
            self.pooling = batch_sum_pooling
        elif pooling == "mean":
            self.pooling = batch_mean_pooling
        else:
            raise NotImplementedError

    def forward(self, graph, x=None, *args):
        batch = graph.batch
        if x is None:
            x = torch.ones((batch.shape[0], 1)).to(batch.device)
        layer_rep = []
        for i in range(self.num_layers):
            x = F.relu(self.bn_layers[i](self.gnn_layers[i](graph, x)))
            layer_rep.append(x)

        pooled_rep = [self.pooling(h, batch) for h in layer_rep]
        node_rep = torch.cat(layer_rep, dim=1)
        graph_rep = torch.cat(pooled_rep, dim=1)
        return graph_rep, node_rep


class FF(nn.Module):
    r"""Residual MLP layers.

    ..math::
        out = \mathbf{MLP}(x) + \mathbf{Linear}(x)

    Paramaters
    ----------
    in_feats : int
        Size of each input sample
    out_feats : int
        Size of each output sample
    """

    def __init__(self, in_feats, out_feats):
        super(FF, self).__init__()
        self.block = MLP(in_feats, out_feats, out_feats, num_layers=3)
        self.shortcut = nn.Linear(in_feats, out_feats)

    def forward(self, x):
        return F.relu(self.block(x)) + self.shortcut(x)


class InfoGraph(BaseModel):
    r"""Implimentation of Infograph in paper `"InfoGraph: Unsupervised and Semi-supervised Graph-Level Representation
     Learning via Mutual Information Maximization" <https://openreview.net/forum?id=r1lfF2NYvH>__. `

     Parameters
     ----------
     in_feats : int
        Size of each input sample.
    out_feats : int
        Size of each output sample.
    num_layers : int, optional
        Number of MLP layers in encoder, default: ``3``.
    unsup : bool, optional
        Use unsupervised model if True, default: ``True``.
    """

    @staticmethod
    def add_args(parser):
        parser.add_argument("--hidden-size", type=int, default=512)
        parser.add_argument("--batch-size", type=int, default=20)
        parser.add_argument("--target", dest="target", type=int, default=0, help="")
        parser.add_argument("--train-num", dest="train_num", type=int, default=5000)
        parser.add_argument("--num-layers", type=int, default=1)
        parser.add_argument("--sup", dest="sup", action="store_true")
        parser.add_argument("--epochs", type=int, default=15)
        parser.add_argument("--lr", type=float, default=0.0001)
        parser.add_argument("--train-ratio", type=float, default=0.7)
        parser.add_argument("--test-ratio", type=float, default=0.1)

    @classmethod
    def build_model_from_args(cls, args):
        return cls(args.num_features, args.hidden_size, args.num_classes, args.num_layers, args.sup)

    @classmethod
    def split_dataset(cls, dataset, args):
        return split_dataset_general(dataset, args)

    def __init__(self, in_feats, hidden_dim, out_feats, num_layers=3, sup=False):
        super(InfoGraph, self).__init__()

        self.sup = sup
        self.emb_dim = hidden_dim
        self.out_feats = out_feats
        self.num_layers = num_layers

        self.sem_fc1 = nn.Linear(num_layers * hidden_dim, hidden_dim)
        self.sem_fc2 = nn.Linear(hidden_dim, out_feats)

        if not sup:
            self.unsup_encoder = Encoder(in_feats, hidden_dim, num_layers)
            self.register_parameter("sem_encoder", None)
        else:
            self.unsup_encoder = Encoder(in_feats, hidden_dim, num_layers)
            self.sem_encoder = Encoder(in_feats, hidden_dim, num_layers)
        self._fc1 = FF(num_layers * hidden_dim, hidden_dim)
        self._fc2 = FF(num_layers * hidden_dim, hidden_dim)
        self.local_dis = FF(num_layers * hidden_dim, hidden_dim)
        self.global_dis = FF(num_layers * hidden_dim, hidden_dim)

        self.criterion = nn.MSELoss()

    def reset_parameters(self):
        for m in self.modules():
            if isinstance(m, nn.Linear):
                nn.init.xavier_uniform_(m.weight.data)

    def forward(self, batch):
        if self.sup:
            return self.sup_forward(batch, batch.x)
        else:
            return self.unsup_forward(batch, batch.x)

    def sup_forward(self, batch, x):
        node_feat, graph_feat = self.sem_encoder(batch, x)
        node_feat = F.relu(self.sem_fc1(node_feat))
        node_feat = self.sem_fc2(node_feat)
        return node_feat

    def unsup_forward(self, batch, x):
        # return self.unsup_loss(x, edge_index, batch)
        graph_feat, node_feat = self.unsup_encoder(batch, x)
        if self.training:
            return graph_feat, node_feat
        else:
            return graph_feat



import torch.nn as nn

from cogdl.layers import HANLayer

from .. import BaseModel


class HAN(BaseModel):
    @staticmethod
    def add_args(parser):
        """Add model-specific arguments to the parser."""
        # fmt: off
        parser.add_argument("--num-features", type=int)
        parser.add_argument("--num-classes", type=int)
        parser.add_argument("--num-nodes", type=int)
        parser.add_argument("--hidden-size", type=int, default=64)
        parser.add_argument("--num-layers", type=int, default=2)
        parser.add_argument("--num-edge", type=int, default=2)
        # fmt: on

    @classmethod
    def build_model_from_args(cls, args):
        return cls(
            args.num_edge,
            args.num_features,
            args.hidden_size,
            args.num_classes,
            args.num_nodes,
            args.num_layers,
        )

    def __init__(self, num_edge, w_in, w_out, num_class, num_nodes, num_layers):
        super(HAN, self).__init__()
        self.num_edge = num_edge
        self.num_nodes = num_nodes
        self.w_in = w_in
        self.w_out = w_out
        self.num_class = num_class
        self.num_layers = num_layers
        layers = []
        for i in range(num_layers):
            if i == 0:
                layers.append(HANLayer(num_edge, w_in, w_out))
            else:
                layers.append(HANLayer(num_edge, w_out, w_out))

        self.layers = nn.ModuleList(layers)
        self.cross_entropy_loss = nn.CrossEntropyLoss()
        self.linear = nn.Linear(self.w_out, self.num_class)

    def forward(self, graph):
        X = graph.x
        for i in range(self.num_layers):
            X = self.layers[i](graph, X)

        out = self.linear(X)
        return out



import copy
import math
from typing import Tuple

from scipy.sparse import linalg

import numpy as np
import scipy.sparse as sparse
import sklearn.preprocessing as preprocessing
import torch
import torch.nn.functional as F

from torch.utils.data import DataLoader

from .. import DataWrapper
from cogdl.data import batch_graphs, Graph


class GCCDataWrapper(DataWrapper):
    @staticmethod
    def add_args(parser):
        # random walk
        parser.add_argument("--batch-size", type=int, default=128)
        parser.add_argument("--rw-hops", type=int, default=64)
        parser.add_argument("--subgraph-size", type=int, default=128)
        parser.add_argument("--restart-prob", type=float, default=0.8)
        parser.add_argument("--positional-embedding-size", type=int, default=128)
        parser.add_argument(
            "--task", type=str, default="node_classification", choices=["node_classification, graph_classification"]
        )
        parser.add_argument("--num-workers", type=int, default=4)

    def __init__(
        self,
        dataset,
        batch_size,
        finetune=False,
        num_workers=4,
        rw_hops=64,
        subgraph_size=128,
        restart_prob=0.8,
        positional_embedding_size=128,
        task="node_classification",
    ):
        super(GCCDataWrapper, self).__init__(dataset)

        data = dataset.data
        data.add_remaining_self_loops()
        if task == "node_classification":
            if finetune:
                self.train_dataset = NodeClassificationDatasetLabeled(
                    data, rw_hops, subgraph_size, restart_prob, positional_embedding_size
                )
            else:
                self.train_dataset = NodeClassificationDataset(
                    data, rw_hops, subgraph_size, restart_prob, positional_embedding_size
                )
        elif task == "graph_classification":
            if finetune:
                pass
            else:
                pass
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.finetune = finetune

        self.rw_hops = rw_hops
        self.subgraph_size = subgraph_size
        self.restart_prob = restart_prob

    def train_wrapper(self):
        train_loader = DataLoader(
            dataset=self.train_dataset,
            batch_size=self.batch_size,
            collate_fn=labeled_batcher() if self.finetune else batcher(),
            shuffle=True if self.finetune else False,
            num_workers=self.num_workers,
            worker_init_fn=None,
        )
        return train_loader


def labeled_batcher():
    def batcher_dev(batch):
        graph_q, label = zip(*batch)
        graph_q = batch_graphs(graph_q)
        return graph_q, torch.LongTensor(label)

    return batcher_dev


def batcher():
    def batcher_dev(batch):
        graph_q_, graph_k_ = zip(*batch)
        graph_q, graph_k = batch_graphs(graph_q_), batch_graphs(graph_k_)
        graph_q.batch_size = len(graph_q_)
        return graph_q, graph_k

    return batcher_dev


def eigen_decomposision(n, k, laplacian, hidden_size, retry):
    if k <= 0:
        return torch.zeros(n, hidden_size)
    laplacian = laplacian.astype("float64")
    ncv = min(n, max(2 * k + 1, 20))
    # follows https://stackoverflow.com/questions/52386942/scipy-sparse-linalg-eigsh-with-fixed-seed
    v0 = np.random.rand(n).astype("float64")
    for i in range(retry):
        try:
            s, u = linalg.eigsh(laplacian, k=k, which="LA", ncv=ncv, v0=v0)
        except sparse.linalg.eigen.arpack.ArpackError:
            # print("arpack error, retry=", i)
            ncv = min(ncv * 2, n)
            if i + 1 == retry:
                sparse.save_npz("arpack_error_sparse_matrix.npz", laplacian)
                u = torch.zeros(n, k)
        else:
            break
    x = preprocessing.normalize(u, norm="l2")
    x = torch.from_numpy(x.astype("float32"))
    x = F.pad(x, (0, hidden_size - k), "constant", 0)
    return x


def _add_undirected_graph_positional_embedding(g: Graph, hidden_size, retry=10):
    # We use eigenvectors of normalized graph laplacian as vertex features.
    # It could be viewed as a generalization of positional embedding in the
    # attention is all you need paper.
    # Recall that the eignvectors of normalized laplacian of a line graph are cos/sin functions.
    # See section 2.4 of http://www.cs.yale.edu/homes/spielman/561/2009/lect02-09.pdf
    n = g.num_nodes
    with g.local_graph():
        g.sym_norm()
        adj = g.to_scipy_csr()
    laplacian = adj

    k = min(n - 2, hidden_size)
    x = eigen_decomposision(n, k, laplacian, hidden_size, retry)
    g.pos_undirected = x.float()
    return g


def _rwr_trace_to_cogdl_graph(
    g: Graph, seed: int, trace: torch.Tensor, positional_embedding_size: int, entire_graph: bool = False
):
    subv = torch.unique(trace).tolist()
    try:
        subv.remove(seed)
    except ValueError:
        pass
    subv = [seed] + subv
    if entire_graph:
        subg = copy.deepcopy(g)
    else:
        subg = g.subgraph(subv)

    subg = _add_undirected_graph_positional_embedding(subg, positional_embedding_size)

    subg.seed = torch.zeros(subg.num_nodes, dtype=torch.long)
    if entire_graph:
        subg.seed[seed] = 1
    else:
        subg.seed[0] = 1
    return subg


class NodeClassificationDataset(object):
    def __init__(
        self,
        data: Graph,
        rw_hops: int = 64,
        subgraph_size: int = 64,
        restart_prob: float = 0.8,
        positional_embedding_size: int = 32,
        step_dist: list = [1.0, 0.0, 0.0],
    ):
        self.rw_hops = rw_hops
        self.subgraph_size = subgraph_size
        self.restart_prob = restart_prob
        self.positional_embedding_size = positional_embedding_size
        self.step_dist = step_dist
        assert positional_embedding_size > 1

        self.data = data
        self.graphs = [self.data]
        self.length = sum([g.num_nodes for g in self.graphs])
        self.total = self.length

    def __len__(self):
        return self.length

    def _convert_idx(self, idx) -> Tuple[int, int]:
        graph_idx = 0
        node_idx = idx
        for i in range(len(self.graphs)):
            if node_idx < self.graphs[i].num_nodes:
                graph_idx = i
                break
            else:
                node_idx -= self.graphs[i].num_nodes
        return graph_idx, node_idx

    def __getitem__(self, idx):
        graph_idx, node_idx = self._convert_idx(idx)

        step = np.random.choice(len(self.step_dist), 1, p=self.step_dist)[0]
        g = self.graphs[graph_idx]

        if step == 0:
            other_node_idx = node_idx
        else:
            other_node_idx = g.random_walk([node_idx], step)[-1]

        max_nodes_per_seed = max(
            self.rw_hops,
            int((self.graphs[graph_idx].degrees()[node_idx] * math.e / (math.e - 1) / self.restart_prob) + 0.5),
        )
        # TODO: `num_workers > 0` is not compatible with `numba`
        traces = g.random_walk_with_restart([node_idx, other_node_idx], max_nodes_per_seed, self.restart_prob)

        # traces = [[0,1,2,3], [1,2,3,4]]

        graph_q = _rwr_trace_to_cogdl_graph(
            g=g,
            seed=node_idx,
            trace=torch.Tensor(traces[0]),
            positional_embedding_size=self.positional_embedding_size,
            entire_graph=hasattr(self, "entire_graph") and self.entire_graph,
        )
        graph_k = _rwr_trace_to_cogdl_graph(
            g=g,
            seed=other_node_idx,
            trace=torch.Tensor(traces[1]),
            positional_embedding_size=self.positional_embedding_size,
            entire_graph=hasattr(self, "entire_graph") and self.entire_graph,
        )
        return graph_q, graph_k


class NodeClassificationDatasetLabeled(NodeClassificationDataset):
    def __init__(
        self,
        data,
        rw_hops=64,
        subgraph_size=64,
        restart_prob=0.8,
        positional_embedding_size=32,
        step_dist=[1.0, 0.0, 0.0],
    ):
        super(NodeClassificationDatasetLabeled, self).__init__(
            data,
            rw_hops,
            subgraph_size,
            restart_prob,
            positional_embedding_size,
            step_dist,
        )
        assert len(self.graphs) == 1
        self.num_classes = self.data.num_classes

    def __getitem__(self, idx):
        graph_idx = 0
        node_idx = idx
        for i in range(len(self.graphs)):
            if node_idx < self.graphs[i].num_nodes:
                graph_idx = i
                break
            else:
                node_idx -= self.graphs[i].num_nodes

        g = self.graphs[graph_idx]
        traces = g.random_walk_with_restart([node_idx], self.rw_hops, self.restart_prob)

        graph_q = _rwr_trace_to_cogdl_graph(
            g=g,
            seed=node_idx,
            trace=torch.Tensor(traces[0]),
            positional_embedding_size=self.positional_embedding_size,
        )
        graph_q.y = self.data.y[idx].y
        return graph_q
        # return graph_q, self.data.y[idx].argmax().item()



import torch


def scatter_add(data, dst, num_nodes, dim=0):
    num_edges, num_feats = data.shape
    out = torch.zeros((num_nodes, num_feats), dtype=data.dtype, device=data.device)
    if len(dst.shape) == 1:
        dst = dst.view(-1, 1)
    index = dst.expand(num_edges, num_feats)
    out = out.scatter_add_(dim=dim, index=index, src=data)
    return out


# [src] - (op) - [edge_attr] - (aggr - op)


def op_src_edge(op, src, e_feat):
    if op == "add":
        return src + e_feat
    elif op == "mul":
        return src * e_feat
    elif op == "sub":
        return src - e_feat
    else:
        raise NotImplementedError


def op_aggr(op, msg, dst, num_nodes):
    if op == "sum":
        out = scatter_add(msg, dst, num_nodes)
    elif op == "mean":
        out = scatter_add(msg, dst, num_nodes)
        deg = torch.zeros((num_nodes,), dtype=torch.float, device=msg.device)
        deg = deg.scatter_add_(dim=0, index=dst, src=torch.ones_like(dst).float())
        deg_inv = deg.pow(-1)
        deg_inv[torch.isinf(deg_inv)] = 0
        out = out * deg_inv.view(-1, 1)
    else:
        raise NotImplementedError
    return out


def src_op_e_aggr_coo(op1, op2, n_feat, e_feat, row, col, data=None):
    nnode = n_feat.shape[0]
    if len(e_feat.shape) == 1:
        e_feat = e_feat.view(-1, 1)
    src = n_feat[col]
    msg = op_src_edge(op1, src, e_feat)
    if data is not None:
        msg = msg * data.view(-1, 1)
    out = op_aggr(op2, msg, row, nnode)
    return out


def s_add_e_sum(g, n_feat, e_feat, weight=False):
    row, col = g.edge_index
    if weight:
        data = g.edge_weight
    else:
        data = None
    return src_op_e_aggr_coo("add", "sum", n_feat, e_feat, row, col, data=data)


def s_mul_e_sum(g, n_feat, e_feat, weight=False):
    row, col = g.edge_index
    if weight:
        data = g.edge_weight
    else:
        data = None
    return src_op_e_aggr_coo("mul", "sum", n_feat, e_feat, row, col, data=data)


def s_sub_e_sum(g, n_feat, e_feat, weight=False):
    row, col = g.edge_index
    if weight:
        data = g.edge_weight
    else:
        data = None
    return src_op_e_aggr_coo("sub", "sum", n_feat, e_feat, row, col, data=data)


def s_add_e_mean(g, n_feat, e_feat, weight=False):
    row, col = g.edge_index
    if weight:
        data = g.edge_weight
    else:
        data = None
    return src_op_e_aggr_coo("add", "mean", n_feat, e_feat, row, col, data=data)


def s_mul_e_mean(g, n_feat, e_feat, weight=False):
    row, col = g.edge_index
    if weight:
        data = g.edge_weight
    else:
        data = None
    return src_op_e_aggr_coo("mul", "mean", n_feat, e_feat, row, col, data=data)


def s_sub_e_mean(g, n_feat, e_feat, weight=False):
    row, col = g.edge_index
    if weight:
        data = g.edge_weight
    else:
        data = None
    return src_op_e_aggr_coo("sub", "mean", n_feat, e_feat, row, col, data=data)


# [src] - (op) - [edge]


def s_add_e(g, n_feat, e_feat):
    _, col = g.edge_index
    return n_feat[col] + e_feat


def s_sub_e(g, n_feat, e_feat):
    _, col = g.edge_index
    return n_feat[col] - e_feat


def s_mul_e(g, n_feat, e_feat):
    _, col = g.edge_index
    return n_feat[col] * e_feat


# [src] - (op) - [dst]


def src_op_target_coo(op, g, src, tgt):
    if tgt is None:
        row, col = g.edge_index
        src, tgt = src[col], src[row]
    if op == "add":
        return src + tgt
    elif op == "mul":
        return src * tgt
    elif op == "sub":
        return src - tgt
    elif op == "dot":
        return (src * tgt).sum(1, keepdim=True)
    elif op == "div":
        out = src / tgt
        out[torch.isinf(out)] = 0
        return out
    else:
        raise NotImplementedError


def s_add_t(g, src, dst=None):
    return src_op_target_coo("add", g, src, dst)


def s_mul_t(g, src, dst=None):
    return src_op_target_coo("mul", g, src, dst)


def s_sub_t(g, src, dst=None):
    return src_op_target_coo("sub", g, src, dst)


def s_dot_t(g, src, dst=None):
    return src_op_target_coo("dot", g, src, dst)


def s_div_t(g, src, dst=None):
    return src_op_target_coo("div", g, src, dst)


def message_passing(send_func, msg_func, rec_fun):
    pass



import numpy as np
import networkx as nx
from gensim.models import Word2Vec, KeyedVectors
import random
from .. import BaseModel


class Metapath2vec(BaseModel):
    r"""The Metapath2vec model from the `"metapath2vec: Scalable Representation
    Learning for Heterogeneous Networks"
    <https://ericdongyx.github.io/papers/
    KDD17-dong-chawla-swami-metapath2vec.pdf>`_ paper

    Args:
        hidden_size (int) : The dimension of node representation.
        walk_length (int) : The walk length.
        walk_num (int) : The number of walks to sample for each node.
        window_size (int) : The actual context size which is considered in language model.
        worker (int) : The number of workers for word2vec.
        iteration (int) : The number of training iteration in word2vec.
        schema (str) : The metapath schema used in model. Metapaths are splited with ",",
        while each node type are connected with "-" in each metapath. For example:"0-1-0,0-2-0,1-0-2-0-1".
    """

    @staticmethod
    def add_args(parser):
        """Add model-specific arguments to the parser."""
        # fmt: off
        parser.add_argument('--walk-length', type=int, default=80,
                            help='Length of walk per source. Default is 80.')
        parser.add_argument('--walk-num', type=int, default=20,
                            help='Number of walks per source. Default is 20.')
        parser.add_argument('--window-size', type=int, default=5,
                            help='Window size of skip-gram model. Default is 5.')
        parser.add_argument('--worker', type=int, default=10,
                            help='Number of parallel workers. Default is 10.')
        parser.add_argument('--iteration', type=int, default=10,
                            help='Number of iterations. Default is 10.')
        parser.add_argument('--schema', type=str, default="No",
                            help="Input schema for multi-type node representation.")
        # fmt: on

    @classmethod
    def build_model_from_args(cls, args):
        return cls(
            args.hidden_size,
            args.walk_length,
            args.walk_num,
            args.window_size,
            args.worker,
            args.iteration,
            args.schema,
        )

    def __init__(self, dimension, walk_length, walk_num, window_size, worker, iteration, schema):
        super(Metapath2vec, self).__init__()
        self.dimension = dimension
        self.walk_length = walk_length
        self.walk_num = walk_num
        self.window_size = window_size
        self.worker = worker
        self.iteration = iteration
        self.schema = schema
        self.node_type = None

    def forward(self, data):
        G = nx.DiGraph()
        row, col = data.edge_index
        G.add_edges_from(list(zip(row.numpy(), col.numpy())))
        self.G = G
        self.node_type = [str(a) for a in data.pos.tolist()]
        walks = self._simulate_walks(self.walk_length, self.walk_num, self.schema)
        walks = [[str(node) for node in walk] for walk in walks]
        model = Word2Vec(
            walks,
            size=self.dimension,
            window=self.window_size,
            min_count=0,
            sg=1,
            workers=self.worker,
            iter=self.iteration,
        )
        id2node = dict([(vid, node) for vid, node in enumerate(G.nodes())])
        embeddings = np.asarray([model.wv[str(id2node[i])] for i in range(len(id2node))])
        return embeddings

    def _walk(self, start_node, walk_length, schema=None):
        # Simulate a random walk starting from start node.
        # Note that metapaths in schema should be like '0-1-0', '0-2-0' or '1-0-2-0-1'.
        if schema:
            schema_items = schema.split("-")
            assert schema_items[0] == schema_items[-1]

        walk = [start_node]
        while len(walk) < walk_length:
            cur = walk[-1]
            candidates = []
            for node in list(self.G.neighbors(cur)):
                if schema is None or self.node_type[node] == schema_items[len(walk) % (len(schema_items) - 1)]:
                    candidates.append(node)
            if candidates:
                walk.append(random.choice(candidates))
            else:
                break
        # print(walk)
        return walk

    def _simulate_walks(self, walk_length, num_walks, schema="No"):
        # Repeatedly simulate random walks from each node with metapath schema.
        G = self.G
        walks = []
        nodes = list(G.nodes())
        if schema != "No":
            schema_list = schema.split(",")
        for walk_iter in range(num_walks):
            random.shuffle(nodes)
            print(str(walk_iter + 1), "/", str(num_walks))
            for node in nodes:
                if schema == "No":
                    walks.append(self._walk(node, walk_length))
                else:
                    for schema_iter in schema_list:
                        if schema_iter.split("-")[0] == self.node_type[node]:
                            walks.append(self._walk(node, walk_length, schema_iter))
        return walks



import torch
from cogdl.wrappers.model_wrapper import ModelWrapper


class HeterogeneousGNNModelWrapper(ModelWrapper):
    def __init__(self, model, optimizer_cfg):
        super(HeterogeneousGNNModelWrapper, self).__init__()
        self.optimizer_cfg = optimizer_cfg
        self.model = model

    def train_step(self, batch):
        graph = batch.data
        pred = self.model(graph)
        train_mask = graph.train_node
        loss = self.default_loss_fn(pred[train_mask], graph.y[train_mask])
        return loss

    def val_step(self, batch):
        graph = batch.data
        pred = self.model(graph)
        val_mask = graph.valid_node
        loss = self.default_loss_fn(pred[val_mask], graph.y[val_mask])
        metric = self.evaluate(pred[val_mask], graph.y[val_mask], metric="auto")
        self.note("val_loss", loss.item())
        self.note("val_metric", metric)

    def test_step(self, batch):
        graph = batch.data
        pred = self.model(graph)
        test_mask = graph.test_node
        loss = self.default_loss_fn(pred[test_mask], graph.y[test_mask])
        metric = self.evaluate(pred[test_mask], graph.y[test_mask], metric="auto")
        self.note("test_loss", loss.item())
        self.note("test_metric", metric)

    def setup_optimizer(self):
        cfg = self.optimizer_cfg
        if hasattr(self.model, "get_optimizer"):
            model_spec_optim = self.model.get_optimizer(cfg)
            if model_spec_optim is not None:
                return model_spec_optim
        return torch.optim.Adam(self.model.parameters(), lr=cfg["lr"], weight_decay=cfg["weight_decay"])



import hashlib

import torch
from joblib import Parallel, delayed
import networkx as nx
import numpy as np
from gensim.models.doc2vec import Doc2Vec, TaggedDocument
import os

from .. import BaseModel


class Graph2Vec(BaseModel):
    r"""The Graph2Vec model from the `"graph2vec: Learning Distributed Representations of Graphs"
    <https://arxiv.org/abs/1707.05005>`_ paper

    Args:
        hidden_size (int) : The dimension of node representation.
        min_count (int) : Parameter in doc2vec.
        window_size (int) : The actual context size which is considered in language model.
        sampling_rate (float) : Parameter in doc2vec.
        dm (int) :  Parameter in doc2vec.
        iteration (int) : The number of iteration in WL method.
        lr (float) : Learning rate in doc2vec.
    """

    @staticmethod
    def add_args(parser):
        parser.add_argument("--hidden-size", type=int, default=128)
        parser.add_argument("--window-size", type=int, default=0)
        parser.add_argument("--min-count", type=int, default=5)
        parser.add_argument("--dm", type=int, default=0)
        parser.add_argument("--sampling", type=float, default=0.0001)
        parser.add_argument("--iteration", type=int, default=2)
        parser.add_argument("--lr", type=float, default=0.025)

    @classmethod
    def build_model_from_args(cls, args):
        return cls(
            args.hidden_size,
            args.min_count,
            args.window_size,
            args.sampling,
            args.dm,
            args.iteration,
            args.epochs,
            args.lr,
        )

    @staticmethod
    def feature_extractor(data, rounds, name):
        edge_index = torch.stack(data.edge_index)
        graph = nx.from_edgelist(np.array(edge_index.T.cpu(), dtype=int))
        if data.x is not None:
            feature = {int(key): str(val) for key, val in enumerate(np.array(data.x.cpu()))}
        else:
            feature = dict(nx.degree(graph))
        graph_wl_features = Graph2Vec.wl_iterations(graph, feature, rounds)
        doc = TaggedDocument(words=graph_wl_features, tags=["g_" + name])
        return doc

    @staticmethod
    def wl_iterations(graph, features, rounds):
        # TODO: solve hash and number
        nodes = graph.nodes
        wl_features = [str(val) for _, val in features.items()]
        for i in range(rounds):
            new_feats = {}
            for node in nodes:
                neighbors = graph.neighbors(node)
                neigh_feats = [features[x] for x in neighbors]
                neigh_feats = [features[node]] + sorted(neigh_feats)
                hash_feat = hashlib.md5("_".join(neigh_feats).encode())
                hash_feat = hash_feat.hexdigest()
                new_feats[node] = hash_feat
            wl_features = wl_features + list(new_feats.values())
            features = new_feats
        return wl_features

    def __init__(self, dimension, min_count, window_size, dm, sampling_rate, rounds, epochs, lr, worker=4):
        super(Graph2Vec, self).__init__()
        self.dimension = dimension
        self.min_count = min_count
        self.window_size = window_size
        self.sampling_rate = sampling_rate
        self.dm = dm
        self.worker = worker
        self.rounds = rounds
        self.model = None
        self.doc_collections = None
        self.epochs = epochs
        self.lr = lr

    def forward(self, graphs, **kwargs):
        if self.doc_collections is None:
            self.doc_collections = Parallel(n_jobs=self.worker)(
                delayed(Graph2Vec.feature_extractor)(graph, self.rounds, str(i)) for i, graph in enumerate(graphs)
            )
        self.model = Doc2Vec(
            self.doc_collections,
            vector_size=self.dimension,
            window=self.window_size,
            min_count=self.min_count,
            dm=self.dm,
            sample=self.sampling_rate,
            workers=self.worker,
            epochs=self.epochs,
            alpha=self.lr,
        )
        vectors = np.array([self.model["g_" + str(i)] for i in range(len(graphs))])
        return vectors

    def save_embedding(self, output_path):
        self.model.wv.save(os.path.join(output_path, "model.wv"))
        self.model.wv.save_word2vec_format(os.path.join("model.emb"))



import torch
import torch.nn as nn
import torch.nn.functional as F

from cogdl.utils import spmm, get_activation, get_norm_layer


class MeanAggregator(object):
    def __call__(self, graph, x):
        graph.row_norm()
        x = spmm(graph, x)
        return x


class SumAggregator(object):
    def __call__(self, graph, x):
        x = spmm(graph, x)
        return x


class MaxAggregator(object):
    def __init__(self):
        from cogdl.operators.scatter_max import scatter_max

        self.scatter_max = scatter_max

    def __call__(self, graph, x):
        x = self.scatter_max(graph.row_indptr.int(), graph.col_indices.int(), x)
        return x


class SAGELayer(nn.Module):
    def __init__(
        self, in_feats, out_feats, normalize=False, aggr="mean", dropout=0.0, norm=None, activation=None, residual=False
    ):
        super(SAGELayer, self).__init__()
        self.in_feats = in_feats
        self.out_feats = out_feats
        self.fc = nn.Linear(2 * in_feats, out_feats)
        self.normalize = normalize
        if dropout > 0:
            self.dropout = nn.Dropout(dropout)
        else:
            self.dropout = None
        if aggr == "mean":
            self.aggr = MeanAggregator()
        elif aggr == "sum":
            self.aggr = SumAggregator()
        elif aggr == "max":
            self.aggr = MaxAggregator()
        else:
            raise NotImplementedError

        if dropout > 0:
            self.dropout = nn.Dropout(dropout)
        else:
            self.dropout = None

        if activation is not None:
            self.act = get_activation(activation, inplace=True)
        else:
            self.act = None

        if norm is not None:
            self.norm = get_norm_layer(norm, out_feats)
        else:
            self.norm = None

        if residual:
            self.residual = nn.Linear(in_features=in_feats, out_features=out_feats)
        else:
            self.residual = None

    def forward(self, graph, x):
        out = self.aggr(graph, x)
        out = torch.cat([x, out], dim=-1)
        out = self.fc(out)
        if self.normalize:
            out = F.normalize(out, p=2.0, dim=-1)

        if self.norm is not None:
            out = self.norm(out)
        if self.act is not None:
            out = self.act(out)

        if self.residual:
            out = out + self.residual(x)

        if self.dropout is not None:
            out = self.dropout(out)

        return out



from typing import Optional, Type, Any
import torch.nn as nn


class BaseModel(nn.Module):
    @staticmethod
    def add_args(parser):
        """Add model-specific arguments to the parser."""
        pass

    @classmethod
    def build_model_from_args(cls, args):
        """Build a new model instance."""
        raise NotImplementedError("Models must implement the build_model_from_args method")

    def __init__(self):
        super(BaseModel, self).__init__()
        self.model_name = self.__class__.__name__
        self.loss_fn = None
        self.evaluator = None

    def _forward_unimplemented(self, *input: Any) -> None:  # abc warning
        pass

    def forward(self, *args):
        raise NotImplementedError

    def predict(self, data):
        return self.forward(data)

    @property
    def device(self):
        return next(self.parameters()).device

    def set_loss_fn(self, loss_fn):
        self.loss_fn = loss_fn



from .embedding_link_prediction_mw import EmbeddingLinkPredictionModelWrapper
from .gnn_kg_link_prediction_mw import GNNKGLinkPredictionModelWrapper
from .gnn_link_prediction_mw import GNNLinkPredictionModelWrapper



import torch
import torch.nn as nn
import torch.nn.functional as F

from .. import BaseModel
from cogdl.layers import MLP
from cogdl.layers import GINLayer
from cogdl.utils import split_dataset_general


class GIN(BaseModel):
    r"""Graph Isomorphism Network from paper `"How Powerful are Graph
    Neural Networks?" <https://arxiv.org/pdf/1810.00826.pdf>`__.

    Args:
        num_layers : int
            Number of GIN layers
        in_feats : int
            Size of each input sample
        out_feats : int
            Size of each output sample
        hidden_dim : int
            Size of each hidden layer dimension
        num_mlp_layers : int
            Number of MLP layers
        eps : float32, optional
            Initial `\epsilon` value, default: ``0``
        pooling : str, optional
            Aggregator type to use, default:　``sum``
        train_eps : bool, optional
            If True, `\epsilon` will be a learnable parameter, default: ``True``
    """

    @staticmethod
    def add_args(parser):
        parser.add_argument("--epsilon", type=float, default=0.0)
        parser.add_argument("--hidden-size", type=int, default=32)
        parser.add_argument("--num-layers", type=int, default=3)
        parser.add_argument("--num-mlp-layers", type=int, default=2)
        parser.add_argument("--dropout", type=float, default=0.5)
        parser.add_argument("--train-epsilon", dest="train_epsilon", action="store_false")
        parser.add_argument("--pooling", type=str, default="sum")

    @classmethod
    def build_model_from_args(cls, args):
        return cls(
            args.num_layers,
            args.num_features,
            args.num_classes,
            args.hidden_size,
            args.num_mlp_layers,
            args.epsilon,
            args.pooling,
            args.train_epsilon,
            args.dropout,
        )

    @classmethod
    def split_dataset(cls, dataset, args):
        return split_dataset_general(dataset, args)

    def __init__(
        self,
        num_layers,
        in_feats,
        out_feats,
        hidden_dim,
        num_mlp_layers,
        eps=0,
        pooling="sum",
        train_eps=False,
        dropout=0.5,
    ):
        super(GIN, self).__init__()
        self.gin_layers = nn.ModuleList()
        self.batch_norm = nn.ModuleList()
        self.num_layers = num_layers
        for i in range(num_layers - 1):
            if i == 0:
                mlp = MLP(in_feats, hidden_dim, hidden_dim, num_mlp_layers, norm="batchnorm")
            else:
                mlp = MLP(hidden_dim, hidden_dim, hidden_dim, num_mlp_layers, norm="batchnorm")
            self.gin_layers.append(GINLayer(mlp, eps, train_eps))
            self.batch_norm.append(nn.BatchNorm1d(hidden_dim))

        self.linear_prediction = nn.ModuleList()
        for i in range(self.num_layers):
            if i == 0:
                self.linear_prediction.append(nn.Linear(in_feats, out_feats))
            else:
                self.linear_prediction.append(nn.Linear(hidden_dim, out_feats))
        self.dropout = nn.Dropout(dropout)
        self.criterion = torch.nn.CrossEntropyLoss()

    def forward(self, batch):
        h = batch.x
        device = h.device
        batchsize = int(torch.max(batch.batch)) + 1

        layer_rep = [h]
        for i in range(self.num_layers - 1):
            h = self.gin_layers[i](batch, h)
            h = self.batch_norm[i](h)
            h = F.relu(h)
            layer_rep.append(h)

        final_score = 0

        for i in range(self.num_layers):
            hsize = layer_rep[i].shape[1]
            output = torch.zeros(batchsize, layer_rep[i].shape[1]).to(device)
            pooled = output.scatter_add_(dim=0, index=batch.batch.view(-1, 1).repeat(1, hsize), src=layer_rep[i])
            final_score += self.dropout(self.linear_prediction[i](pooled))
        return final_score



import numpy as np

from .. import DataWrapper
from cogdl.wrappers.tools.wrapper_utils import node_degree_as_feature


class GraphEmbeddingDataWrapper(DataWrapper):
    @staticmethod
    def add_args(parser):
        # fmt: off
        parser.add_argument("--degree-node-features", action="store_true",
                            help="Use one-hot degree vector as input node features")
        # fmt: on

    def __init__(self, dataset, degree_node_features=False):
        super(GraphEmbeddingDataWrapper, self).__init__(dataset)
        self.dataset = dataset
        self.degree_node_features = degree_node_features

    def train_wrapper(self):
        return self.dataset

    def test_wrapper(self):
        if self.dataset[0].y.shape[0] > 1:
            return np.array([g.y.numpy() for g in self.dataset])
        else:
            return np.array([g.y.item() for g in self.dataset])

    def pre_transform(self):
        if self.degree_node_features:
            self.dataset = node_degree_as_feature(self.dataset)



import copy

import torch
import numpy as np
from sklearn.cluster import KMeans

from cogdl.wrappers.data_wrapper import DataWrapper

from .node_classification_mw import NodeClfModelWrapper


class M3SModelWrapper(NodeClfModelWrapper):
    @staticmethod
    def add_args(parser):
        # fmt: off
        parser.add_argument("--n-cluster", type=int, default=10)
        parser.add_argument("--num-new-labels", type=int, default=10)
        # fmt: on

    def __init__(self, model, optimizer_cfg, n_cluster, num_new_labels):
        super(M3SModelWrapper, self).__init__(model, optimizer_cfg)
        self.model = model
        self.num_clusters = n_cluster
        self.hidden_size = optimizer_cfg["hidden_size"]
        self.num_new_labels = num_new_labels
        self.optimizer_cfg = optimizer_cfg

    def pre_stage(self, stage, data_w: DataWrapper):
        if stage > 0:
            graph = data_w.get_dataset().data
            graph.store("y")

            y = copy.deepcopy(graph.y)

            num_classes = graph.num_classes
            num_nodes = graph.num_nodes

            with torch.no_grad():
                emb = self.model.embed(graph)

            confidence_ranking = np.zeros([num_classes, num_nodes], dtype=int)
            kmeans = KMeans(n_clusters=self.num_clusters, random_state=0).fit(emb)
            clusters = kmeans.labels_

            # Compute centroids μ_m of each class m in labeled data and v_l of each cluster l in unlabeled data.
            labeled_centroid = np.zeros([num_classes, self.hidden_size])
            unlabeled_centroid = np.zeros([self.num_clusters, self.hidden_size])
            for i in range(num_nodes):
                if graph.train_mask[i]:
                    labeled_centroid[y[i]] += emb[i]
                else:
                    unlabeled_centroid[clusters[i]] += emb[i]

            # Align labels for each cluster
            align = np.zeros(self.num_clusters, dtype=int)
            for i in range(self.num_clusters):
                for j in range(num_classes):
                    if np.linalg.norm(unlabeled_centroid[i] - labeled_centroid[j]) < np.linalg.norm(
                        unlabeled_centroid[i] - labeled_centroid[align[i]]
                    ):
                        align[i] = j

            # Add new labels
            for i in range(num_classes):
                t = self.num_new_labels
                for j in range(num_nodes):
                    idx = confidence_ranking[i][j]
                    if not graph.train_mask[idx]:
                        if t <= 0:
                            break
                        t -= 1
                        if align[clusters[idx]] == i:
                            graph.train_mask[idx] = True
                            y[idx] = i
            return y



import optuna
import numpy as np
import random

from .. import BaseModel
from cogdl.utils.prone_utils import propagate, get_embedding_dense


class PlainFilter(object):
    def __init__(self, filter_types, svd):
        self.filter_types = filter_types
        self.svd = svd
        # load adjacency matrix and raw embedding

    def __call__(self, emb, adj):
        if len(self.filter_types) == 1 and self.filter_types[0] == "identity":
            return emb

        dim = emb.shape[1]
        prop_result = []
        for tp in self.filter_types:
            prop_result.append(propagate(adj, emb, tp))
        prop_result_emb = np.concatenate(prop_result, axis=1)
        if self.svd:
            prop_result_emb = get_embedding_dense(prop_result_emb, dim)
        return prop_result_emb


class Search(object):
    def __init__(self, filter_types, max_evals, svd, loss_type, n_workers):
        self.prop_types = filter_types
        self.max_evals = max_evals
        self.svd = svd
        self.loss_type = loss_type
        self.n_workers = n_workers

        # load adjacency matrix and raw embedding
        self.num_edges = self.num_nodes = self.dim = 0
        self.laplacian = None
        self.emb = self.adj = None

        self.batch_size = 64

    def build_search_space(self, trial):
        space = {}
        for f in self.prop_types:
            space[f] = trial.suggest_categorical(f, [0, 1])
        if space.get("heat", 0) == 1:
            space["t"] = trial.suggest_uniform("t", 0.1, 0.9)
        if space.get("gaussian", 0) == 1:
            space["mu"] = trial.suggest_uniform("mu", 0.1, 2)
            space["theta"] = trial.suggest_uniform("theta", 0.2, 1.5)
        if space.get("ppr", 0) == 1:
            space["alpha"] = trial.suggest_uniform("alpha", 0.2, 0.8)
        return space

    def init_data(self, emb, adj):
        self.num_nodes, self.dim = emb.shape
        self.num_edges = adj.nnz
        self.emb = emb
        self.adj = adj

        if self.loss_type == "infonce":
            neg_index = []
            for i in range(self.num_nodes):
                select = np.random.choice(self.num_nodes, self.batch_size, replace=False)
                while i in select:
                    select = np.random.choice(self.num_nodes, self.batch_size, replace=False)
                neg_index.append(select)
            self.neg_index = np.array(neg_index)
            self.neg_emb = self.emb[self.neg_index]
        elif self.loss_type == "infomax":
            # self.permutation = np.random.permutation(np.arange(self.num_nodes))
            pass

    def prop(self, params):
        prop_types = [key for key, value in params.items() if value == 1 and key in self.prop_types]
        if not prop_types:
            print(" -- dropped -- ")
            return None, None

        prop_result_list = []
        for selected_prop in prop_types:
            prop_result = propagate(self.adj, self.emb, selected_prop, params)
            prop_result_list.append(prop_result)

        if self.loss_type == "infomax":
            neg_prop_result = []
            pmt = self.permutation
            for s_prop in prop_types:
                neg_prop = propagate(self.adj, self.emb[pmt], s_prop, params)
                neg_prop[pmt] = neg_prop
                neg_prop_result.append(neg_prop)
            return np.array(prop_result_list), np.array(neg_prop_result)
        elif self.loss_type == "infonce":
            return np.array(prop_result_list), None
        elif self.loss_type == "sparse":
            return np.array(prop_result_list), None
        else:
            raise ValueError("use 'infonce', 'infomax' or 'sparse' loss, currently using {}".format(self.loss_type))

    def target_func(self, trial):
        params = self.build_search_space(trial)
        self.permutation = np.random.permutation(np.arange(self.num_nodes))
        prop_result_emb, neg_prop_result_emb = self.prop(params)
        if prop_result_emb is None:
            return 100
        if self.loss_type == "infomax":
            loss = self.infomax_loss(prop_result_emb, neg_prop_result_emb)
        elif self.loss_type == "infonce":
            loss = self.infonce_loss(prop_result_emb)
        else:
            raise ValueError("loss type must be in ['infomax', 'infonce']")
        return loss

    def infonce_loss(self, prop_emb_list, *args, **kwargs):
        T = 0.07

        pos_infos = []
        for smoothed in prop_emb_list:
            pos_info = np.exp(np.sum(smoothed * self.emb, -1) / T)
            assert pos_info.shape == (self.num_nodes,)
            pos_infos.append(pos_info)

        neg_infos = []
        for idx, smoothed in enumerate(prop_emb_list):
            neg_info = np.exp(
                np.sum(np.tile(smoothed[:, np.newaxis, :], (1, self.batch_size, 1)) * self.neg_emb, -1) / T
            ).sum(-1)
            assert neg_info.shape == (self.num_nodes,)
            neg_infos.append(neg_info + pos_infos[idx])

        pos_neg = np.array(pos_infos) / np.array(neg_infos)
        if np.isnan(pos_neg).any():
            pos_neg = np.nan_to_num(pos_neg)

        loss = -np.log(pos_neg).mean()
        return loss / 10

    def infomax_loss(self, prop_emb_list, neg_prop_emb_list):
        prop_result = np.concatenate(prop_emb_list, axis=1)
        if self.svd:
            prop_result = get_embedding_dense(prop_result, self.dim)

        def sigmoid(x):
            return 1.0 / (1 + np.exp(-x))

        pos_glb = prop_result.mean(0)
        pos_info = sigmoid(pos_glb.dot(prop_result.T))
        pos_loss = np.mean(np.log(pos_info)).mean()

        neg_loss = 0
        neg_step = 1
        for _ in range(neg_step):
            neg_prop_result = np.concatenate(neg_prop_emb_list, axis=1)
            if self.svd:
                neg_prop_result = get_embedding_dense(neg_prop_result, self.dim)

            neg_info = sigmoid(pos_glb.dot(neg_prop_result.T))
            neg_loss += np.mean(np.log(1 - neg_info)).mean()
            random.shuffle(neg_prop_emb_list)

        return -(pos_loss + neg_loss) / (1 + neg_step)

    def __call__(self, emb, adj):
        self.init_data(emb, adj)
        study = optuna.create_study()
        study.optimize(self.target_func, n_jobs=self.n_workers, n_trials=self.max_evals)
        best_params = study.best_params

        best_result = self.prop(best_params)[0]
        best_result = np.concatenate(best_result, axis=1)
        print(f"best parameters: {best_params}")

        if self.svd:
            best_result = get_embedding_dense(best_result, self.dim)
        return best_result


class ProNEPP(BaseModel):
    @staticmethod
    def add_args(parser):
        pass

    @classmethod
    def build_model_from_args(cls, args):
        return cls(
            filter_types=args.filter_types,
            max_evals=args.max_evals,
            loss_type=args.loss,
            svd=not args.no_svd,
            n_workers=args.num_workers,
            search=not args.no_search,
        )

    def __init__(self, filter_types, svd, search, max_evals=None, loss_type=None, n_workers=None):
        super(ProNEPP, self).__init__()
        if search:
            self.model = Search(filter_types, max_evals, svd, loss_type, n_workers)
        else:
            self.model = PlainFilter(filter_types, svd)

    def __call__(self, emb, adj):
        enhanced_emb = self.model(emb, adj)
        return enhanced_emb



import torch

try:
    import fused_gatconv

    def fused_gat_func(attn_row, attn_col, row_ptr, col_ind, col_ptr, row_ind, negative_slope, in_feat):
        return FusedGATFunction.apply(attn_row, attn_col, row_ptr, col_ind, col_ptr, row_ind, negative_slope, in_feat)


except Exception:
    fused_gat_func = None


class FusedGATFunction(torch.autograd.Function):
    @staticmethod
    def forward(ctx, attn_row, attn_col, row_ptr, col_ind, col_ptr, row_ind, negative_slope, in_feat):
        out_feat, edge_max, edge_sum = fused_gatconv.gat_forward(
            attn_row, attn_col, row_ptr, col_ind, negative_slope, in_feat
        )
        ctx.save_for_backward(row_ptr, col_ind, col_ptr, row_ind, edge_max, edge_sum, in_feat, attn_row, attn_col)
        ctx.negative_slope = negative_slope
        return out_feat

    @staticmethod
    def backward(ctx, grad_out):
        row_ptr, col_ind, col_ptr, row_ind, edge_max, edge_sum, in_feat, attn_row, attn_col = ctx.saved_tensors
        grad_out = grad_out.contiguous()
        grad_feat, grad_attn_row, grad_attn_col = fused_gatconv.gat_backward(
            ctx.negative_slope,
            row_ptr,
            col_ind,
            col_ptr,
            row_ind,
            edge_max,
            edge_sum,
            in_feat,
            attn_row,
            attn_col,
            grad_out,
        )
        return grad_attn_row, grad_attn_col, None, None, None, None, None, grad_feat, None



import unicodedata
from tqdm import tqdm

COLORCODES = {
    "black": "\x1b[30m",
    "red": "\x1b[31m",
    "green": "\x1b[32m",
    "yellow": "\x1b[33m",
    "blue": "\x1b[34m",
    "magenta": "\x1b[35m",
    "cyan": "\x1b[36m",
    "white": "\x1b[37m",
    "reset": "\x1b[0m",
}


def colored(text, color):
    return COLORCODES.get(color, "") + text + COLORCODES.get("reset", "")


OAG_TOKEN_TYPE_NAMES = ["TEXT", "AUTHOR", "VENUE", "AFF", "FOS", "FUND"]


def stringLenCJK(string):
    return sum(1 + (unicodedata.east_asian_width(c) in "WF") for c in string)


def stringRjustCJK(string, length):
    return " " * (length - stringLenCJK(string)) + string



import os
import torch

from ogb.nodeproppred import NodePropPredDataset
from ogb.nodeproppred import Evaluator as NodeEvaluator
from ogb.graphproppred import GraphPropPredDataset

from cogdl.data import Dataset, Graph, DataLoader
from cogdl.utils import CrossEntropyLoss, Accuracy, remove_self_loops, coalesce, BCEWithLogitsLoss


class OGBNDataset(Dataset):
    def __init__(self, root, name, transform=None):
        name = name.replace("-", "_")
        self.name = name
        root = os.path.join(root, name)
        super(OGBNDataset, self).__init__(root)
        self.transform = None
        self.data = torch.load(self.processed_paths[0])

    def get(self, idx):
        assert idx == 0
        return self.data

    def get_loss_fn(self):
        return CrossEntropyLoss()

    def get_evaluator(self):
        return Accuracy()

    def _download(self):
        pass

    @property
    def processed_file_names(self):
        return "data_cogdl.pt"

    def process(self):
        name = self.name.replace("_", "-")
        dataset = NodePropPredDataset(name, self.root)
        graph, y = dataset[0]
        x = torch.tensor(graph["node_feat"]).contiguous() if graph["node_feat"] is not None else None
        y = torch.tensor(y.squeeze())
        row, col = graph["edge_index"][0], graph["edge_index"][1]
        row = torch.from_numpy(row)
        col = torch.from_numpy(col)
        edge_index = torch.stack([row, col], dim=0)
        edge_attr = torch.as_tensor(graph["edge_feat"]) if graph["edge_feat"] is not None else graph["edge_feat"]
        edge_index, edge_attr = remove_self_loops(edge_index, edge_attr)
        row = torch.cat([edge_index[0], edge_index[1]])
        col = torch.cat([edge_index[1], edge_index[0]])

        row, col, _ = coalesce(row, col)
        edge_index = torch.stack([row, col], dim=0)

        data = Graph(x=x, edge_index=edge_index, edge_attr=edge_attr, y=y)
        data.num_nodes = graph["num_nodes"]

        # split
        split_index = dataset.get_idx_split()
        data.train_mask = torch.full((data.num_nodes,), False, dtype=torch.bool)
        data.val_mask = torch.full((data.num_nodes,), False, dtype=torch.bool)
        data.test_mask = torch.full((data.num_nodes,), False, dtype=torch.bool)

        data.train_mask[split_index["train"]] = True
        data.test_mask[split_index["test"]] = True
        data.val_mask[split_index["valid"]] = True

        torch.save(data, self.processed_paths[0])
        return data


class OGBArxivDataset(OGBNDataset):
    def __init__(self, data_path="data"):
        dataset = "ogbn-arxiv"
        super(OGBArxivDataset, self).__init__(data_path, dataset)


class OGBProductsDataset(OGBNDataset):
    def __init__(self, data_path="data"):
        dataset = "ogbn-products"
        super(OGBProductsDataset, self).__init__(data_path, dataset)


class OGBProteinsDataset(OGBNDataset):
    def __init__(self, data_path="data"):
        dataset = "ogbn-proteins"
        super(OGBProteinsDataset, self).__init__(data_path, dataset)

    @property
    def edge_attr_size(self):
        return [
            self.data.edge_attr.shape[1],
        ]

    def get_loss_fn(self):
        return BCEWithLogitsLoss()

    def get_evaluator(self):
        evaluator = NodeEvaluator(name="ogbn-proteins")

        def wrap(y_pred, y_true):
            input_dict = {"y_true": y_true, "y_pred": y_pred}
            return evaluator.eval(input_dict)["rocauc"]

        return wrap

    def process(self):
        name = self.name.replace("_", "-")
        dataset = NodePropPredDataset(name, self.root)
        graph, y = dataset[0]
        y = torch.tensor(y.squeeze())
        row, col = graph["edge_index"][0], graph["edge_index"][1]
        row = torch.from_numpy(row)
        col = torch.from_numpy(col)
        edge_attr = torch.as_tensor(graph["edge_feat"]) if "edge_feat" in graph else None

        data = Graph(x=None, edge_index=(row, col), edge_attr=edge_attr, y=y)
        data.num_nodes = graph["num_nodes"]

        # split
        split_index = dataset.get_idx_split()
        data.train_mask = torch.full((data.num_nodes,), False, dtype=torch.bool)
        data.val_mask = torch.full((data.num_nodes,), False, dtype=torch.bool)
        data.test_mask = torch.full((data.num_nodes,), False, dtype=torch.bool)

        data.train_mask[split_index["train"]] = True
        data.test_mask[split_index["test"]] = True
        data.val_mask[split_index["valid"]] = True

        edge_attr = data.edge_attr
        deg = data.degrees()
        dst, _ = data.edge_index
        dst = dst.view(-1, 1).expand(dst.shape[0], edge_attr.shape[1])
        x = torch.zeros((data.num_nodes, edge_attr.shape[1]), dtype=torch.float32)
        x = x.scatter_add_(dim=0, index=dst, src=edge_attr)
        deg = torch.clamp(deg, min=1)
        x = x / deg.view(-1, 1)
        data.x = x

        node_species = torch.as_tensor(graph["node_species"])
        n_species, new_index = torch.unique(node_species, return_inverse=True)
        one_hot_x = torch.nn.functional.one_hot(new_index, num_classes=torch.max(new_index).int().item())
        data.species = node_species
        data.x = torch.cat([data.x, one_hot_x], dim=1)
        torch.save(data, self.processed_paths[0])
        return data


class OGBPapers100MDataset(OGBNDataset):
    def __init__(self, data_path="data"):
        dataset = "ogbn-papers100M"
        super(OGBPapers100MDataset, self).__init__(data_path, dataset)


class OGBGDataset(Dataset):
    def __init__(self, root, name):
        super(OGBGDataset, self).__init__(root)
        self.name = name
        self.dataset = GraphPropPredDataset(self.name, root)

        self.data = []
        self.all_nodes = 0
        self.all_edges = 0
        for i in range(len(self.dataset.graphs)):
            graph, label = self.dataset[i]
            data = Graph(
                x=torch.tensor(graph["node_feat"], dtype=torch.float),
                edge_index=torch.tensor(graph["edge_index"]),
                edge_attr=None if "edge_feat" not in graph else torch.tensor(graph["edge_feat"], dtype=torch.float),
                y=torch.tensor(label),
            )
            data.num_nodes = graph["num_nodes"]
            self.data.append(data)

            self.all_nodes += graph["num_nodes"]
            self.all_edges += graph["edge_index"].shape[1]

        self.transform = None

    def get_loader(self, args):
        split_index = self.dataset.get_idx_split()
        train_loader = DataLoader(self.get_subset(split_index["train"]), batch_size=args.batch_size, shuffle=True)
        valid_loader = DataLoader(self.get_subset(split_index["valid"]), batch_size=args.batch_size, shuffle=False)
        test_loader = DataLoader(self.get_subset(split_index["test"]), batch_size=args.batch_size, shuffle=False)
        return train_loader, valid_loader, test_loader

    def get_subset(self, subset):
        datalist = []
        for idx in subset:
            datalist.append(self.data[idx])
        return datalist

    def get(self, idx):
        return self.data[idx]

    def _download(self):
        pass

    def _process(self):
        pass

    @property
    def num_classes(self):
        return int(self.dataset.num_classes)


class OGBMolbaceDataset(OGBGDataset):
    def __init__(self, data_path="data"):
        dataset = "ogbg-molbace"
        super(OGBMolbaceDataset, self).__init__(data_path, dataset)


class OGBMolhivDataset(OGBGDataset):
    def __init__(self, data_path="data"):
        dataset = "ogbg-molhiv"
        super(OGBMolhivDataset, self).__init__(data_path, dataset)


class OGBMolpcbaDataset(OGBGDataset):
    def __init__(self, data_path="data"):
        dataset = "ogbg-molpcba"
        super(OGBMolpcbaDataset, self).__init__(data_path, dataset)


class OGBPpaDataset(OGBGDataset):
    def __init__(self):
        dataset = "ogbg-ppa"
        path = "data"
        super(OGBPpaDataset, self).__init__(path, dataset)


class OGBCodeDataset(OGBGDataset):
    def __init__(self, data_path="data"):
        dataset = "ogbg-code"
        super(OGBCodeDataset, self).__init__(data_path, dataset)
